from supervisely.app.widgets import Widget, DynamicWidget
from supervisely.app.widgets_context import JinjaWidgets
from supervisely.app.fastapi.websocket import WebsocketManager
from supervisely.app.fastapi.utils import run_sync
from supervisely.app.content import DataJson, StateJson


class ReloadableArea(DynamicWidget):
    """Widget for dynamic content reloading. It allows to update the content of the widget without reloading the whole
    UI of the app. It can be used when the widgets are needed to be added or removed dynamically.

    :param content: a Widget to be set as content of the ReloadableArea. Defaults to None.
    :type content: Widget, optional
    :param widget_id: The id of the widget. Defaults to None.
    :type widget_id: str, optional

    :Methods:
    reload(): Reloads the widget in UI.
    set_content(): Replaces content of the ReloadableArea with new widget.

    :Usage example:

     .. code-block:: python
        from supervisely.app.widgets import ReloadableArea, Container, Button

        # Creating button, which will be added to widget at initialization.
        button_plus = Button("plus")
        buttons_container = Container(widgets=[button_plus])

        # Initializing ReloadableArea with buttons_container as content.
        reloadable_area = ReloadableArea(content=buttons_container)

        # Now we need to create new widget, after UI was initialized.
        button_minus = Button("minus")
        buttons_container._widgets.append(button_minus)

        # Widget of new button was added to container, but it will not appear in UI
        # until we reload the ReloadableArea.

        reloadable_area.reload()

        # Now the new button will appear in UI.
    """

    def __init__(self, content: Widget = None, widget_id: str = None):
        self._content = content
        super().__init__(widget_id=widget_id, file_path=__file__)

        script_path = "./sly/css/app/widgets/reloadable_area/script.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__] = script_path

    def get_json_data(self):
        return None

    def get_json_state(self):
        return None

    def set_content(self, content: Widget):
        """Replaces content of the ReloadableArea with new widget. Note: this method doesn't reload the widget.
        To reload the widget in UI use reload() function.

        :param content: new widget to be set as content
        :type content: Widget
        """
        self._content = content
        DataJson().send_changes()
        StateJson().send_changes()

    def reload(self):
        """Reloads the widget in UI."""
        DataJson().send_changes()
        StateJson().send_changes()

        html_content = self._content.to_html()
        run_sync(
            WebsocketManager().broadcast(
                {
                    "runAction": {
                        "action": f"rerender-template-{self.widget_id}",
                        "payload": {"template": html_content},
                    }
                }
            )
        )

    def hide(self):
        """Hides the content of the ReloadableArea."""
        self._content.hide()
        
    def show(self):
        """Shows the content of the ReloadableArea."""
        self._content.show()



# coding: utf-8
"""get list of :class:`objects<supervisely.annotation.obj_class.ObjClass>` from supervisely project"""

from __future__ import annotations

from typing import Dict, List, NamedTuple, Optional

from supervisely.api.module_api import ApiField, ModuleApi
from supervisely.video_annotation.key_id_map import KeyIdMap


class ObjectClassApi(ModuleApi):
    """
    API for working with :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>`. :class:`ObjectClassApi<ObjectClassApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        project_id = 1951
        obj_class_infos = api.object_class.get_list(project_id)
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple ObjectClassInfo information about ObjectClass.

        :Example:

         .. code-block:: python

            ObjectClassInfo(id=22309,
                            name='lemon',
                            description='',
                            shape='bitmap',
                            color='#51C6AA',
                            settings={},
                            created_at='2021-03-02T10:04:33.973Z',
                            updated_at='2021-03-11T09:37:07.111Z')
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.DESCRIPTION,
            ApiField.SHAPE,
            ApiField.COLOR,
            ApiField.SETTINGS,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **ObjectClassInfo**.
        """
        return "ObjectClassInfo"

    def get_list(
        self, project_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[NamedTuple]:
        """
        List of ObjClasses in the given Project.

        :param project_id: Project ID in which the ObjClasses are located.
        :type project_id: int
        :param filters: List of params to sort output ObjClasses.
        :type filters: List[dict], optional
        :return: List of ObjClasses with information from the given Project. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[NamedTuple]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_id = 1951
            obj_class_infos = api.object_class.get_list(project_id)
            print(obj_class_infos)
            # Output: [ObjectClassInfo(id=22309,
            #                          name='lemon',
            #                          description='',
            #                          shape='bitmap',
            #                          color='#51C6AA',
            #                          settings={},
            #                          created_at='2021-03-02T10:04:33.973Z',
            #                          updated_at='2021-03-11T09:37:07.111Z'),
            #  ObjectClassInfo(id=22310,
            #                  name='kiwi',
            #                  description='',
            #                  shape='bitmap',
            #                  color='#FF0000',
            #                  settings={},
            #                  created_at='2021-03-02T10:04:33.973Z',
            #                  updated_at='2021-03-11T09:37:07.111Z')
            # ]

            obj_class_list = api.object_class.get_list(1951, filters=[{'field': 'name', 'operator': '=', 'value': 'lemon' }])
            print(obj_class_list)
            # Output: [
            #     [
            #         22309,
            #         "lemon",
            #         "",
            #         "bitmap",
            #         "#51C6AA",
            #         {},
            #         "2021-03-02T10:04:33.973Z",
            #         "2021-03-11T09:37:07.111Z"
            #     ]
            # ]
        """
        return self.get_list_all_pages(
            "advanced.object_classes.list",
            {ApiField.PROJECT_ID: project_id, "filter": filters or []},
        )

    def get_name_to_id_map(self, project_id: int) -> Dict[str, int]:
        """
        :param project_id: Project ID in which the ObjClasses are located.
        :type project_id: int
        :return: Dictionary Key ID Map {'key': id}
        :rtype: :class:`KeyIdMap<supervisely.video_annotation.key_id_map.KeyIdMap>`
        :Usage example:

        .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            obj_class_map = api.object_class.get_name_to_id_map(1951)
            print(obj_class_map)
            # Output: {'lemon': 22309, 'kiwi': 22310, 'cucumber': 22379}
        """
        objects_infos = self.get_list(project_id)
        return {object_info.name: object_info.id for object_info in objects_infos}

    def _get_info_by_id(self, id, method, fields=None):
        response = self._get_response_by_id(id, method, id_field=ApiField.ID, fields=fields)
        return (
            self._convert_json_info(response.json(), skip_missing=True)
            if (response is not None)
            else None
        )

    def get_info_by_id(self, id):
        return self._get_info_by_id(
            id,
            "advanced.object_classes.info",
        )

    # def _object_classes_to_json(self, object_classes: KeyIndexedCollection, objclasses_name_id_map=None, project_id=None):
    #     pass #@TODO: implement
    #     # if objclasses_name_id_map is None and project_id is None:
    #     #     raise RuntimeError("Impossible to get ids for projectTags")
    #     # if objclasses_name_id_map is None:
    #     #     objclasses_name_id_map = self.get_name_to_id_map(project_id)
    #     # tags_json = []
    #     # for tag in tags:
    #     #     tag_json = tag.to_json()
    #     #     tag_json[ApiField.TAG_ID] = tag_name_id_map[tag.name]
    #     #     tags_json.append(tag_json)
    #     # return tags_json
    #
    # def append_to_video(self, video_id, tags: KeyIndexedCollection, key_id_map: KeyIdMap = None):
    #     if len(tags) == 0:
    #         return []
    #     video_info = self._api.video.get_info_by_id(video_id)
    #     tags_json = self._tags_to_json(tags, project_id=video_info.project_id)
    #     ids = self.append_to_video_json(video_id, tags_json)
    #     KeyIdMap.add_tags_to(key_id_map, [tag.key() for tag in tags], ids)
    #     return ids
    #
    # def append_to_video_json(self, video_id, tags_json):
    #     if len(tags_json) == 0:
    #         return
    #     response = self._api.post('videos.tags.bulk.add', {ApiField.VIDEO_ID: video_id, ApiField.TAGS: tags_json})
    #     ids = [obj[ApiField.ID] for obj in response.json()]
    #     return ids

    def update(
        self,
        id: int,
        name: Optional[str] = None,
        description: Optional[str] = None,
        hotkey: Optional[str] = None,
        shape: Optional[str] = None,
        color: Optional[str] = None,
        settings: Optional[dict] = None,
    ) -> NamedTuple:
        """
        Update the class with the given ID on the server.
        Returned object contains updated information about the class except settings.

        :param id: ID of the class to update.
        :type id: int
        :param name: New name of the class.
        :type name: str, optional
        :param description: New description of the class.
        :type description: str, optional
        :param hotkey: New hotkey of the class (e.g., "K").
        :type hotkey: str, optional
        :param shape: New shape of the class.
        :type shape: str, optional
        :param color: New color of the class in HEX format (e.g., #FFFFFF).
        :type color: str, optional
        :param settings: New settings of the class.
                        Do not pass "availableShapes" for shape other than "any".
                        Do not pass "availableShapes" that does not contain the current shape.
        :type settings: dict, optional

        :return: Updated class information
        :rtype: :class:`ObjectClassInfo<ObjectClassInfo>`

        :Usage example:

        .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()

            obj_class_info = api.object_class.update(
                id=22309,
                shape='any',
                settings={
                    "availableShapes": [
                        "bitmap",
                        "polygon",
                    ],
                },
            )
        """
        if all(arg is None for arg in [name, description, hotkey, shape, color, settings]):
            raise ValueError(
                f"To update the class with ID: {id}, you must specify at least one parameter to update; all are currently None"
            )

        data = {ApiField.ID: id}
        if name is not None:
            data[ApiField.NAME] = name
        if description is not None:
            data[ApiField.DESCRIPTION] = description
        if hotkey is not None:
            data[ApiField.HOTKEY] = hotkey
        if shape is not None:
            data[ApiField.SHAPE] = shape
        if color is not None:
            data[ApiField.COLOR] = color
        if settings is not None:
            data[ApiField.SETTINGS] = settings

        response = self._api.post("advanced.object_classes.editInfo", data)
        return self._convert_json_info(response.json(), skip_missing=True)



from typing import Union
from functools import wraps
from supervisely.app.widgets.apexchart.apexchart import Apexchart
from supervisely.app.content import StateJson, DataJson

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

"""
chart = sly.app.widgets.HeatmapChart(
    title="Objects count distribution for every class",
    xaxis_title="Number of objects on image",
    color_range="row",
    tooltip="There are {y} images with {x} objects of class {series_name}",
)
"""


class HeatmapChart(Apexchart):
    def __init__(
        self,
        title: str,
        data_labels: bool = True,
        xaxis_title: str = None,
        color_range: Literal["table", "row"] = "row",
        tooltip: str = None,
    ):
        self._title = title
        self._series = []
        self._original_series_x = {}
        self._data_labels = data_labels
        self._xaxis_title = xaxis_title
        self._widget_height = 350
        self._color_range = color_range
        self._distributed = False
        self._colors = ["#008FFB"]
        self._tooltip = tooltip
        if self._color_range == "row":
            self._distributed = True
            self._colors = [
                "#F3B415",
                "#F27036",
                "#663F59",
                "#6A6E94",
                "#4E88B4",
                "#00A7C6",
                "#18D8D8",
                "#A9D794",
                "#46AF78",
                "#A93F55",
                "#8C5E58",
                "#2176FF",
                "#33A1FD",
                "#7A918D",
                # "#BAFF29",
                "#bce95c",
            ]

        self._options = {
            "chart": {
                "type": "heatmap",
                "toolbar": {
                    "tools": {
                        "download": True,
                        "selection": False,
                        "zoom": False,
                        "zoomin": False,
                        "zoomout": False,
                        "pan": False,
                        "reset": False,
                    },
                },
            },
            "legend": {"show": False},
            "plotOptions": {
                "heatmap": {
                    "distributed": self._distributed,
                    "colorScale": {
                        "ranges": [
                            {"from": 0, "to": 0, "name": "", "color": "#DCDCDC"},
                        ]
                    },
                },
            },
            "dataLabels": {"enabled": self._data_labels, "style": {"colors": ["#fff"]}},
            "colors": self._colors,
            "title": {"text": self._title, "align": "left"},
            "xaxis": {},
        }
        if self._xaxis_title is not None:
            self._options["xaxis"]["title"] = {"text": str(self._xaxis_title)}

        sly_options = {}
        if self._tooltip is not None:
            sly_options["tooltip"] = self._tooltip
            # self._options["tooltip"] = {"followCursor": False}
            self._options["tooltip"] = {"y": {}}

        super(HeatmapChart, self).__init__(
            series=self._series,
            options=self._options,
            type="heatmap",
            height=self._widget_height,
            sly_options=sly_options,
        )

    def _update_height(self):
        self._widget_height = len(self._series) * 40 + 70
        DataJson()[self.widget_id]["height"] = self._widget_height

    def add_series_batch(self, series: dict):
        # usage example
        # lines = []
        # for class_name, x, y in stats.get_series():
        #     lines.append({"name": class_name, "x": x, "y": y})
        for serie in series:
            name = serie["name"]
            x = serie["x"]
            y = serie["y"]
            self.add_series(name, x, y, send_changes=False)
        DataJson().send_changes()

    def add_series(self, name: str, x: list, y: list, send_changes=True):
        x_str = [str(num) for num in x]
        self._original_series_x[name] = x.copy()
        super().add_series(name, x_str, y, send_changes)
        self._update_height()

    def get_clicked_datapoint(self):
        res = super().get_clicked_datapoint()
        if res is not None:
            res = res._replace(x=self._original_series_x[res.series_name][res.data_index])
        return res



from typing import Dict, List, Tuple, Union

# pylint: disable=import-error
import clip
import numpy as np

from supervisely import Annotation, Label, VideoAnnotation
from supervisely.nn.tracker.deep_sort import generate_clip_detections as gdet
from supervisely.nn.tracker.deep_sort import preprocessing
from supervisely.nn.tracker.deep_sort.deep_sort import nn_matching
from supervisely.nn.tracker.deep_sort.deep_sort.detection import (
    Detection as dsDetection,
)
from supervisely.nn.tracker.deep_sort.deep_sort.track import Track as dsTrack
from supervisely.nn.tracker.deep_sort.deep_sort.track import TrackState
from supervisely.nn.tracker.deep_sort.deep_sort.tracker import Tracker as dsTracker
from supervisely.nn.tracker.tracker import BaseDetection, BaseTrack, BaseTracker
from supervisely.sly_logger import logger


class Detection(BaseDetection, dsDetection):
    def __init__(self, sly_label: Label, tlwh, confidence, feature):
        dsDetection.__init__(self, tlwh, confidence, feature)
        self.sly_label = sly_label

    def get_sly_label(self):
        return self.sly_label


class Track(BaseTrack, dsTrack):
    def __init__(
        self,
        mean,
        covariance,
        track_id,
        n_init,
        max_age,
        detection: Detection = None,
    ):
        dsTrack.__init__(self, mean, covariance, track_id, n_init, max_age, feature=None)

        self.state = TrackState.Confirmed
        self.features = []
        self._sly_label = None
        self.class_num = None
        if detection is not None:
            self.features.append(detection.feature)
            self._sly_label = detection.get_sly_label()
            self.class_num = self._sly_label.obj_class.name

    def get_sly_label(self):
        return self._sly_label

    def clean_sly_label(self):
        self._sly_label = None

    def update(self, kf, detection: Detection):
        dsTrack.update(self, kf, detection)
        self._sly_label = detection.get_sly_label()


class _dsTracker(dsTracker):
    """Extend deep sort tracker to support Supervisely labels."""

    def update(self, detections: List[Detection]):
        """Perform measurement update and track management.

        Parameters
        ----------
        detections : List[deep_sort.detection.Detection]
            A list of detections at the current time step.

        """
        # Clean up all previous labels
        for track in self.tracks:
            track: Track
            track.clean_sly_label()

        dsTracker.update(self, detections)

    def _initiate_track(self, detection: Detection):
        mean, covariance = self.kf.initiate(detection.to_xyah())
        self.tracks.append(
            Track(mean, covariance, self._next_id, self.n_init, self.max_age, detection)
        )
        self._next_id += 1


class DeepSortTracker(BaseTracker):
    def __init__(self, settings: Dict = None):
        if settings is None:
            settings = {}
        super().__init__(settings)
        model_filename = "ViT-B/32"  # initialize deep sort
        logger.info("Loading CLIP...")
        model, transform = clip.load(model_filename, device=self.device)
        self.encoder = gdet.create_box_encoder(model, transform, batch_size=1, device=self.device)
        metric = nn_matching.NearestNeighborDistanceMetric(  # calculate cosine distance metric
            "cosine", self.args.max_cosine_distance, self.args.nn_budget
        )
        self.tracker = _dsTracker(metric, n_init=1)

    def default_settings(self):
        """To be overridden by subclasses."""
        return {"nms_max_overlap": 1.0, "max_cosine_distance": 0.6, "nn_budget": None}

    def track(
        self,
        source: Union[List[np.ndarray], List[str], str],
        frame_to_annotation: Dict[int, Annotation],
        frame_shape: Tuple[int, int],
        pbar_cb=None,
    ) -> VideoAnnotation:
        """
        Track objects in the video using DeepSort algorithm.

        :param source: List of images, paths to images or path to the video file.
        :type source: List[np.ndarray] | List[str] | str
        :param frame_to_annotation: Dictionary with frame index as key and Annotation as value.
        :type frame_to_annotation: Dict[int, Annotation]
        :param frame_shape: Size of the frame (height, width).
        :type frame_shape: Tuple[int, int]
        :param pbar_cb: Callback to update progress bar.
        :type pbar_cb: Callable, optional

        :return: Video annotation with tracked objects.
        :rtype: VideoAnnotation

        :raises ValueError: If number of images and annotations are not the same.

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.nn.tracker import DeepSortTracker

            api = sly.Api()

            project_id = 12345
            video_id = 12345678
            video_path = "video.mp4"

            # Download video and get video info
            video_info = api.video.get_info_by_id(video_id)
            frame_shape = (video_info.frame_height, video_info.frame_width)
            api.video.download_path(id=video_id, path=video_path)

            # Run inference app to get detections
            task_id = 12345 # detection app task id
            session = sly.nn.inference.Session(api, task_id)
            annotations = session.inference_video_id(video_id, 0, video_info.frames_count)
            frame_to_annotation = {i: ann for i, ann in enumerate(annotations)}

            # Run tracker
            tracker = DeepSortTracker()
            video_ann = tracker.track(video_path, frame_to_annotation, frame_shape)

            # Upload result
            model_meta = session.get_model_meta()
            project_meta = sly.ProjectMeta.from_json(api.project.get_meta(project_id))
            project_meta = project_meta.merge(model_meta)
            api.project.update_meta(project_id, project_meta)
            api.video.annotation.append(video_id, video_ann)
        """
        if not isinstance(source, str):
            if len(source) != len(frame_to_annotation):
                raise ValueError("Number of images and annotations should be the same")

        tracks_data = {}
        logger.info("Starting deep_sort tracking with CLIP...")

        for frame_index, img in enumerate(self.frames_generator(source)):
            tracks_data = self.update(
                img, frame_to_annotation[frame_index], frame_index, tracks_data
            )

            if pbar_cb is not None:
                pbar_cb()

        tracks_data = self.clear_empty_ids(tracker_annotations=tracks_data)

        return self.get_annotation(
            tracks_data=tracks_data,
            frame_shape=frame_shape,
            frames_count=len(frame_to_annotation),
        )

    def update(
        self, img, annotation: Annotation, frame_index, tracks_data: Dict[int, List[Dict]] = None
    ):
        import torch

        detections = []
        try:
            pred, sly_labels = self.convert_annotation(annotation)
            det = torch.tensor(pred)

            # Process detections
            bboxes = det[:, :4].clone().cpu()
            # tlwh -> lthw
            bboxes = [bbox[[1, 0, 3, 2]] for bbox in bboxes]
            confs = det[:, 4]

            # encode yolo detections and feed to tracker
            features = self.encoder(img, bboxes)
            detections = [
                Detection(sly_label, bbox, conf, feature)
                for bbox, conf, feature, sly_label in zip(bboxes, confs, features, sly_labels)
            ]

            # run non-maxima supression
            boxs = np.array([d.tlwh for d in detections])
            scores = np.array([d.confidence for d in detections])
            class_nums = np.array([d.sly_label.obj_class.name for d in detections])
            indices_of_alive_labels = preprocessing.non_max_suppression(
                boxs, class_nums, self.args.nms_max_overlap, scores
            )
            detections = [detections[i] for i in indices_of_alive_labels]
        except Exception as ex:
            import traceback

            logger.info(f"frame {frame_index} skipped on tracking")
            logger.debug(traceback.format_exc())

        # Call the tracker
        self.tracker.predict()
        self.tracker.update(detections)

        if tracks_data is None:
            tracks_data = {}
        self.update_track_data(
            tracks_data=tracks_data,
            tracks=[
                track
                for track in self.tracker.tracks
                if track.is_confirmed() or track.time_since_update <= 1
            ],
            frame_index=frame_index,
        )
        return tracks_data

    def update_track_data(self, tracks_data: dict, tracks: List[BaseTrack], frame_index: int):
        track_id_data = []
        labels_data = []

        for curr_track in tracks:
            track_id = curr_track.track_id - 1  # track_id starts from 1

            if curr_track.get_sly_label() is not None:
                track_id_data.append(track_id)
                labels_data.append(curr_track.get_sly_label())

        tracks_data[frame_index] = {"ids": track_id_data, "labels": labels_data}

        return tracks_data

    def clear_empty_ids(self, tracker_annotations):
        id_mappings = {}
        last_ordinal_id = 0

        for frame_index, data in tracker_annotations.items():
            data_ids_temp = []
            for current_id in data["ids"]:
                new_id = id_mappings.get(current_id, -1)
                if new_id == -1:
                    id_mappings[current_id] = last_ordinal_id
                    last_ordinal_id += 1
                    new_id = id_mappings.get(current_id, -1)
                data_ids_temp.append(new_id)
            data["ids"] = data_ids_temp

        return tracker_annotations



import time
import uuid
from queue import Queue
from threading import Event, Thread
from typing import BinaryIO, Dict, List, Tuple

import numpy as np
from pydantic import ValidationError

from supervisely._utils import find_value_by_keys
from supervisely.annotation.label import Geometry, Label
from supervisely.annotation.obj_class import ObjClass
from supervisely.api.api import Api
from supervisely.api.module_api import ApiField
from supervisely.api.video.video_figure_api import FigureInfo
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.helpers import deserialize_geometry
from supervisely.geometry.polygon import Polygon
from supervisely.imaging import image as sly_image
from supervisely.nn.inference.tracking.base_tracking import BaseTracking
from supervisely.nn.inference.tracking.tracker_interface import (
    TrackerInterface,
    TrackerInterfaceV2,
)
from supervisely.sly_logger import logger
from supervisely.task.progress import Progress


class MaskTracking(BaseTracking):
    def _deserialize_geometry(self, data: dict):
        geometry_type_str = data["type"]
        geometry_json = data["data"]
        return deserialize_geometry(geometry_type_str, geometry_json)

    def _inference(self, frames: List[np.ndarray], geometries: List[Geometry]):
        results = [[] for _ in range(len(frames) - 1)]
        for geometry in geometries:
            if not isinstance(geometry, Bitmap) and not isinstance(geometry, Polygon):
                raise TypeError(f"This app does not support {geometry.geometry_name()} tracking")

            # combine several binary masks into one multilabel mask
            i = 0
            label2id = {}

            original_geometry = geometry.clone()

            # convert polygon to bitmap
            if isinstance(geometry, Polygon):
                polygon_obj_class = ObjClass("polygon", Polygon)
                polygon_label = Label(geometry, polygon_obj_class)
                bitmap_obj_class = ObjClass("bitmap", Bitmap)
                bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                geometry = bitmap_label.geometry
            if i == 0:
                multilabel_mask = geometry.data.astype(int)
                multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
                geometry.draw(bitmap=multilabel_mask, color=[1, 1, 1])
                i += 1
            else:
                i += 1
                geometry.draw(bitmap=multilabel_mask, color=[i, i, i])
            label2id[i] = {
                "original_geometry": original_geometry.geometry_name(),
            }
            # run tracker
            tracked_multilabel_masks = self.predict(
                frames=frames, input_mask=multilabel_mask[:, :, 0]
            )
            tracked_multilabel_masks = np.array(tracked_multilabel_masks)
            # decompose multilabel masks into binary masks
            for i in np.unique(tracked_multilabel_masks):
                if i != 0:
                    binary_masks = tracked_multilabel_masks == i
                    geometry_type = label2id[i]["original_geometry"]
                    for j, mask in enumerate(binary_masks[1:]):
                        # check if mask is not empty
                        if np.any(mask):
                            if geometry_type == "polygon":
                                bitmap_geometry = Bitmap(mask)
                                bitmap_obj_class = ObjClass("bitmap", Bitmap)
                                bitmap_label = Label(bitmap_geometry, bitmap_obj_class)
                                polygon_obj_class = ObjClass("polygon", Polygon)
                                polygon_labels = bitmap_label.convert(polygon_obj_class)
                                geometries = [label.geometry for label in polygon_labels]
                            else:
                                geometries = [Bitmap(mask)]
                            results[j].extend(
                                [
                                    {"type": geom.geometry_name(), "data": geom.to_json()}
                                    for geom in geometries
                                ]
                            )
        return results

    def _track(self, api: Api, context: Dict):
        self.video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=True,
            notify_in_predict=True,
            per_point_polygon_tracking=False,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
        )
        range_of_frames = [
            self.video_interface.frames_indexes[0],
            self.video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            # if cache is persistent, run cache task for whole video
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=self.video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=self.video_interface.video_id,
            )

        api.logger.info("Starting tracking process")
        # load frames
        frames = self.video_interface.frames
        # combine several binary masks into one multilabel mask
        i = 0
        label2id = {}

        def _upload_loop(q: Queue, stop_event: Event, video_interface: TrackerInterface):
            try:
                while True:
                    items = []
                    while not q.empty():
                        items.append(q.get_nowait())
                    if len(items) > 0:
                        video_interface.add_object_geometries_on_frames(*list(zip(*items)))
                        continue
                    if stop_event.is_set():
                        video_interface._notify(True, task="stop tracking")
                        return
                    time.sleep(1)
            except Exception as e:
                api.logger.error("Error in upload loop: %s", str(e), exc_info=True)
                video_interface._notify(True, task="stop tracking")
                video_interface.global_stop_indicatior = True
                raise

        upload_queue = Queue()
        stop_upload_event = Event()
        Thread(
            target=_upload_loop,
            args=[upload_queue, stop_upload_event, self.video_interface],
            daemon=True,
        ).start()

        try:
            for (fig_id, geometry), obj_id in zip(
                self.video_interface.geometries.items(),
                self.video_interface.object_ids,
            ):
                original_geometry = geometry.clone()
                if not isinstance(geometry, Bitmap) and not isinstance(geometry, Polygon):
                    stop_upload_event.set()
                    raise TypeError(
                        f"This app does not support {geometry.geometry_name()} tracking"
                    )
                # convert polygon to bitmap
                if isinstance(geometry, Polygon):
                    polygon_obj_class = ObjClass("polygon", Polygon)
                    polygon_label = Label(geometry, polygon_obj_class)
                    bitmap_obj_class = ObjClass("bitmap", Bitmap)
                    bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                    geometry = bitmap_label.geometry
                if i == 0:
                    multilabel_mask = geometry.data.astype(int)
                    multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
                    geometry.draw(bitmap=multilabel_mask, color=[1, 1, 1])
                    i += 1
                else:
                    i += 1
                    geometry.draw(bitmap=multilabel_mask, color=[i, i, i])
                label2id[i] = {
                    "fig_id": fig_id,
                    "obj_id": obj_id,
                    "original_geometry": original_geometry.geometry_name(),
                }
            # run tracker
            tracked_multilabel_masks = self.predict(
                frames=frames, input_mask=multilabel_mask[:, :, 0]
            )
            tracked_multilabel_masks = np.array(tracked_multilabel_masks)
            # decompose multilabel masks into binary masks
            for i in np.unique(tracked_multilabel_masks):
                if i != 0:
                    binary_masks = tracked_multilabel_masks == i
                    fig_id = label2id[i]["fig_id"]
                    obj_id = label2id[i]["obj_id"]
                    geometry_type = label2id[i]["original_geometry"]
                    for j, mask in enumerate(binary_masks[1:]):
                        # check if mask is not empty
                        if not np.any(mask):
                            api.logger.info(
                                f"Skipping empty mask on frame {self.video_interface.frame_index + j + 1}"
                            )
                            # update progress bar anyway (otherwise it will not be finished)
                            self.video_interface._notify(task="add geometry on frame")
                        else:
                            if geometry_type == "polygon":
                                bitmap_geometry = Bitmap(mask)
                                bitmap_obj_class = ObjClass("bitmap", Bitmap)
                                bitmap_label = Label(bitmap_geometry, bitmap_obj_class)
                                polygon_obj_class = ObjClass("polygon", Polygon)
                                polygon_labels = bitmap_label.convert(polygon_obj_class)
                                geometries = [label.geometry for label in polygon_labels]
                            else:
                                geometries = [Bitmap(mask)]
                            for l, geometry in enumerate(geometries):
                                if l == len(geometries) - 1:
                                    notify = True
                                else:
                                    notify = False
                                upload_queue.put(
                                    (
                                        geometry,
                                        obj_id,
                                        self.video_interface.frames_indexes[j + 1],
                                        notify,
                                    )
                                )
                    if self.video_interface.global_stop_indicatior:
                        stop_upload_event.set()
                        return
                    api.logger.info(f"Figure with id {fig_id} was successfully tracked")
        except Exception:
            stop_upload_event.set()
            raise
        stop_upload_event.set()

    def _track_async(self, api: Api, context: dict, inference_request_uuid: str = None):
        inference_request = self._inference_requests[inference_request_uuid]
        tracker_interface = TrackerInterfaceV2(api, context, self.cache)
        progress: Progress = inference_request["progress"]
        frames_count = tracker_interface.frames_count
        figures = tracker_interface.figures
        progress_total = frames_count * len(figures)
        progress.total = progress_total

        def _upload_f(items: List[Tuple[FigureInfo, bool]]):
            with inference_request["lock"]:
                inference_request["pending_results"].extend([item[0] for item in items])

        def _notify_f(items: List[Tuple[FigureInfo, bool]]):
            items_by_object_id: Dict[int, List[Tuple[FigureInfo, bool]]] = {}
            for item in items:
                items_by_object_id.setdefault(item[0].object_id, []).append(item)

            for object_id, object_items in items_by_object_id.items():
                frame_range = [
                    min(item[0].frame_index for item in object_items),
                    max(item[0].frame_index for item in object_items),
                ]
                progress.iters_done_report(sum(1 for item in object_items if item[1]))
                tracker_interface.notify_progress(progress.current, progress.total, frame_range)

        # run tracker
        frame_index = tracker_interface.frame_index
        direction_n = tracker_interface.direction_n
        api.logger.info("Start tracking.")
        try:
            with tracker_interface(_upload_f, _notify_f):
                # combine several binary masks into one multilabel mask
                i = 0
                label2id = {}
                # load frames
                frames = tracker_interface.load_all_frames()
                frames = [frame.image for frame in frames]
                for figure in figures:
                    figure = api.video.figure._convert_json_info(figure)
                    fig_id = figure.id
                    obj_id = figure.object_id
                    geometry = deserialize_geometry(figure.geometry_type, figure.geometry)
                    original_geometry = geometry.clone()
                    if not isinstance(geometry, (Bitmap, Polygon)):
                        raise TypeError(
                            f"This app does not support {geometry.geometry_name()} tracking"
                        )
                    # convert polygon to bitmap
                    if isinstance(geometry, Polygon):
                        polygon_obj_class = ObjClass("polygon", Polygon)
                        polygon_label = Label(geometry, polygon_obj_class)
                        bitmap_obj_class = ObjClass("bitmap", Bitmap)
                        bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                        geometry = bitmap_label.geometry
                    if i == 0:
                        multilabel_mask = geometry.data.astype(int)
                        multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
                        geometry.draw(bitmap=multilabel_mask, color=[1, 1, 1])
                        i += 1
                    else:
                        i += 1
                        geometry.draw(bitmap=multilabel_mask, color=[i, i, i])
                    label2id[i] = {
                        "fig_id": fig_id,
                        "obj_id": obj_id,
                        "original_geometry": original_geometry.geometry_name(),
                    }
                    if inference_request["cancel_inference"]:
                        return
                    if tracker_interface.is_stopped():
                        reason = tracker_interface.stop_reason()
                        if isinstance(reason, Exception):
                            raise reason
                        return

                # predict
                tracked_multilabel_masks = self.predict(
                    frames=frames, input_mask=multilabel_mask[:, :, 0]
                )
                tracked_multilabel_masks = np.array(tracked_multilabel_masks)

                # decompose multilabel masks into binary masks
                for i in np.unique(tracked_multilabel_masks):
                    if inference_request["cancel_inference"]:
                        return
                    if tracker_interface.is_stopped():
                        reason = tracker_interface.stop_reason()
                        if isinstance(reason, Exception):
                            raise reason
                        return
                    if i != 0:
                        binary_masks = tracked_multilabel_masks == i
                        fig_id = label2id[i]["fig_id"]
                        obj_id = label2id[i]["obj_id"]
                        geometry_type = label2id[i]["original_geometry"]
                        for j, mask in enumerate(binary_masks[1:], 1):
                            if inference_request["cancel_inference"]:
                                return
                            if tracker_interface.is_stopped():
                                reason = tracker_interface.stop_reason()
                                if isinstance(reason, Exception):
                                    raise reason
                                return
                            this_figure_index = frame_index + j * direction_n
                            # check if mask is not empty
                            if not np.any(mask):
                                api.logger.info(f"Skipping empty mask on frame {this_figure_index}")
                                # update progress bar anyway (otherwise it will not be finished)
                                progress.iter_done_report()
                            else:
                                if geometry_type == "polygon":
                                    bitmap_geometry = Bitmap(mask)
                                    bitmap_obj_class = ObjClass("bitmap", Bitmap)
                                    bitmap_label = Label(bitmap_geometry, bitmap_obj_class)
                                    polygon_obj_class = ObjClass("polygon", Polygon)
                                    polygon_labels = bitmap_label.convert(polygon_obj_class)
                                    geometries = [label.geometry for label in polygon_labels]
                                else:
                                    geometries = [Bitmap(mask)]
                                for l, geometry in enumerate(geometries):
                                    figure_id = uuid.uuid5(
                                        namespace=uuid.NAMESPACE_URL, name=f"{time.time()}"
                                    ).hex
                                    result_figure = api.video.figure._convert_json_info(
                                        {
                                            ApiField.ID: figure_id,
                                            ApiField.OBJECT_ID: obj_id,
                                            "meta": {"frame": this_figure_index},
                                            ApiField.GEOMETRY_TYPE: geometry.geometry_name(),
                                            ApiField.GEOMETRY: geometry.to_json(),
                                            ApiField.TRACK_ID: tracker_interface.track_id,
                                        }
                                    )
                                    should_notify = l == len(geometries) - 1
                                    tracker_interface.add_prediction((result_figure, should_notify))
                        api.logger.info(
                            "Figure [%d, %d] tracked.",
                            i,
                            len(figures),
                            extra={"figure_id": figure.id},
                        )
        except Exception:
            progress.message = "Error occured during tracking"
            raise
        else:
            progress.message = "Ready"
        finally:
            progress.set(current=0, total=1, report=True)

    # Implement the following methods in the derived class
    def track(self, api: Api, state: Dict, context: Dict):
        fn = self.send_error_data(api, context)(self._track)
        self.schedule_task(fn, api, context)
        return {"message": "Tracking has started."}

    def track_api(self, api: Api, state: Dict, context: Dict):
        # unused fields:
        context["trackId"] = "auto"
        context["objectIds"] = []
        context["figureIds"] = []
        if "direction" not in context:
            context["direction"] = "forward"

        input_geometries: list = context["input_geometries"]

        self.video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=True,
            notify_in_predict=True,
            per_point_polygon_tracking=False,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
            should_notify=False,
        )

        range_of_frames = [
            self.video_interface.frames_indexes[0],
            self.video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            # if cache is persistent, run cache task for whole video
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=self.video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=self.video_interface.video_id,
            )

        api.logger.info("Starting tracking process")
        # load frames
        frames = self.video_interface.frames
        # combine several binary masks into one multilabel mask
        i = 0
        label2id = {}

        for input_geom in input_geometries:
            geometry = self._deserialize_geometry(input_geom)
            if not isinstance(geometry, Bitmap) and not isinstance(geometry, Polygon):
                raise TypeError(f"This app does not support {geometry.geometry_name()} tracking")
            # convert polygon to bitmap
            if isinstance(geometry, Polygon):
                polygon_obj_class = ObjClass("polygon", Polygon)
                polygon_label = Label(geometry, polygon_obj_class)
                bitmap_obj_class = ObjClass("bitmap", Bitmap)
                bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                geometry = bitmap_label.geometry
            if i == 0:
                multilabel_mask = geometry.data.astype(int)
                multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
                geometry.draw(bitmap=multilabel_mask, color=[1, 1, 1])
                i += 1
            else:
                i += 1
                geometry.draw(bitmap=multilabel_mask, color=[i, i, i])
            label2id[i] = {
                "original_geometry": geometry.geometry_name(),
            }

        # run tracker
        tracked_multilabel_masks = self.predict(frames=frames, input_mask=multilabel_mask[:, :, 0])
        tracked_multilabel_masks = np.array(tracked_multilabel_masks)

        predictions = []
        # decompose multilabel masks into binary masks
        for i in np.unique(tracked_multilabel_masks):
            if i != 0:
                predictions_for_label = []
                binary_masks = tracked_multilabel_masks == i
                geometry_type = label2id[i]["original_geometry"]
                for j, mask in enumerate(binary_masks[1:]):
                    # check if mask is not empty
                    if not np.any(mask):
                        api.logger.info(f"Empty mask on frame {context['frameIndex'] + j + 1}")
                        predictions_for_label.append(None)
                    else:
                        # frame_idx = j + 1
                        geometry = Bitmap(mask)
                        predictions_for_label.append(
                            {"type": geometry.geometry_name(), "data": geometry.to_json()}
                        )
                predictions.append(predictions_for_label)

        # predictions must be NxK masks: N=number of frames, K=number of objects
        predictions = list(map(list, zip(*predictions)))
        return predictions

    def track_api_files(
        self,
        files: List[BinaryIO],
        settings: Dict,
    ):
        logger.info(f"Start tracking with settings:", extra={"settings": settings})
        frame_index = find_value_by_keys(settings, ["frame_index", "frameIndex"])
        frames_count = find_value_by_keys(settings, ["frames", "framesCount"])
        input_geometries = find_value_by_keys(settings, ["input_geometries", "inputGeometries"])
        direction = settings.get("direction", "forward")
        direction_n = 1 if direction == "forward" else -1
        frame_indexes = list(
            range(frame_index, frame_index + frames_count * direction_n + direction_n, direction_n)
        )
        geometries = map(self._deserialize_geometry, input_geometries)
        frames = []
        for file, frame_idx in zip(files, frame_indexes):
            img_bytes = file.read()
            frame = sly_image.read_bytes(img_bytes)
            frames.append(frame)
        logger.info("Start tracking.")
        return self._inference(frames, geometries)

    def track_async(self, api: Api, state: Dict, context: Dict):
        batch_size = context.get("batch_size", self.get_batch_size())
        if self.max_batch_size is not None and batch_size > self.max_batch_size:
            raise ValidationError(
                f"Batch size should be less than or equal to {self.max_batch_size} for this model."
            )

        inference_request_uuid = uuid.uuid5(namespace=uuid.NAMESPACE_URL, name=f"{time.time()}").hex
        fn = self.send_error_data(api, context)(self._track_async)
        self.schedule_task(fn, api, context, inference_request_uuid=inference_request_uuid)

        logger.debug(
            "Inference has scheduled from 'track_async' endpoint",
            extra={"inference_request_uuid": inference_request_uuid},
        )
        return {
            "message": "Inference has started.",
            "inference_request_uuid": inference_request_uuid,
        }



from typing import Callable, Dict, List, Optional, Union

import supervisely.io.env as env
from supervisely.api.api import Api
from supervisely.app.widgets import Widget
from supervisely.app.widgets.checkbox.checkbox import Checkbox
from supervisely.app.widgets.container.container import Container
from supervisely.app.widgets.select.select import Select
from supervisely.app.widgets.tree_select.tree_select import TreeSelect
from supervisely.project.project_type import ProjectType


class SelectDatasetTree(Widget):
    """SelectDatasetTree widget in Supervisely is a widget that allows users to select datasets from a tree-like structure.
    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/selection/selectdatasettree>`_
        (including screenshots and examples).

    :param default_id: The ID of the dataset to be selected by default.
    :type default_id: Union[int, None]
    :param project_id: The ID of the project to read datasets from.
    :type project_id: Union[int, None]
    :param multiselect: Whether multiple datasets can be selected.
    :type multiselect: bool
    :param compact: Whether the widget should be compact (e.g. no team, workspace, and project selectors).
    :type compact: bool
    :param select_all_datasets: Whether all datasets should be selected by default.
    :type select_all_datasets: bool
    :param allowed_project_types: The list of project types that are allowed to be selected.
    :type allowed_project_types: Optional[List[ProjectType]]
    :param flat: Whether the dataset selector should be flat.
    :type flat: bool
    :param always_open: Whether the dataset selector should always be open.
    :type always_open: bool
    :param team_is_selectable: Whether the team selector should be selectable.
    :type team_is_selectable: bool
    :param workspace_is_selectable: Whether the workspace selector should be selectable.
    :type workspace_is_selectable: bool
    :param append_to_body: Determines where the popover is attached. If False, it is positioned inside the input's container. This can cause the popover to be hidden if the input is within a Card or a widget that restricts visibility.
    :type append_to_body: bool
    :param widget_id: The unique identifier of the widget.
    :type widget_id: Union[str, None]
    :param show_select_all_datasets_checkbox: Whether the checkbox to select all datasets should be shown.
    :type show_select_all_datasets_checkbox: bool

    :Public methods:
    - `set_project_id(project_id: int) -> None`: Set the project ID to read datasets from.
    - `get_selected_ids() -> Optional[List[int]]`: Get the IDs of the selected datasets.
    - `get_selected_id() -> Optional[int]`: Get the ID of the selected dataset.
    - `value_changed(func: Callable) -> Callable`: Decorator to set the callback function for the value changed event.
    - `set_dataset_id(dataset_id: int) -> None`: Set the ID of the dataset to be selected by default.
    - `set_dataset_ids(dataset_ids: List[int]) -> None`: Set the IDs of the datasets to be selected by default.
    - `get_selected_project_id() -> Optional[int]`: Get the ID of the selected project.
    - `get_selected_team_id() -> int`: Get the ID of the selected team.
    - `set_team_id(team_id: int) -> None`: Set the team ID to read workspaces from.
    - `get_selected_workspace_id() -> int`: Get the ID of the selected workspace.
    - `set_workspace_id(workspace_id: int) -> None`: Set the workspace ID to read projects from.
    - `is_all_selected() -> bool`: Check if all datasets are selected.
    - `select_all() -> None`: Select all datasets.

    :Properties:
    - `team_id`: The ID of the team selected in the widget.
    - `workspace_id`: The ID of the workspace selected in the widget.
    - `project_id`: The ID of the project selected in the widget.

    :Usage example:

        .. code-block:: python
            from supervisely.app.widgets import SelectDatasetTree

            project_id = 123
            dataset_id = 456

            select_dataset_tree = SelectDatasetTree(
                default_id=dataset_id,
                project_id=project_id,
                multiselect=True,
                flat=True)

            @select_dataset_tree.value_changed
            def on_change(selected_ids):
                print(selected_ids) # Output: [456, 789]
    """

    def __init__(
        self,
        default_id: Union[int, None] = None,
        project_id: Union[int, None] = None,
        multiselect: bool = False,
        compact: bool = False,
        select_all_datasets: bool = False,
        allowed_project_types: Optional[List[ProjectType]] = None,
        flat: bool = False,
        always_open: bool = False,
        team_is_selectable: bool = True,
        workspace_is_selectable: bool = True,
        append_to_body: bool = True,
        widget_id: Union[str, None] = None,
        show_select_all_datasets_checkbox: bool = True,
        width: int = 193,
    ):
        self._api = Api.from_env()

        if default_id is not None and project_id is None:
            raise ValueError("Project ID must be provided when default dataset ID is set.")

        if not multiselect and select_all_datasets:
            raise ValueError("Select all datasets is only available in multiselect mode.")

        # Reading team_id and workspace_id from environment variables.
        # If not found, error will be raised.
        self._team_id = env.team_id()
        self._workspace_id = env.workspace_id()

        # Using environment variables to set the default values if they are not provided.
        self._project_id = project_id or env.project_id(raise_not_found=False)
        self._dataset_id = default_id or env.dataset_id(raise_not_found=False)

        self._multiselect = multiselect
        self._compact = compact
        self._append_to_body = append_to_body

        # Extract values from Enum to match the .type property of the ProjectInfo object.

        self._project_types = None
        if allowed_project_types is not None:
            if all(allowed_project_types) is isinstance(allowed_project_types, ProjectType):
                self._project_types = (
                    [project_type.value for project_type in allowed_project_types]
                    if allowed_project_types is not None
                    else None
                )
            elif all(allowed_project_types) is isinstance(allowed_project_types, str):
                self._project_types = allowed_project_types

        # Widget components.
        self._select_team = None
        self._select_workspace = None
        self._select_project = None
        self._select_dataset = None
        self._width = width

        # List of widgets will be used to create a Container.
        self._widgets = []

        if not compact:
            # If the widget is not compact, create team, workspace, and project selectors.
            self._create_selectors(team_is_selectable, workspace_is_selectable)

        # Create the dataset selector.
        self._create_dataset_selector(flat, always_open, select_all_datasets)

        # Create the checkbox to select all datasets if needed.
        self._select_all_datasets_checkbox = None
        if show_select_all_datasets_checkbox:
            self._create_select_all_datasets_checkbox(select_all_datasets)

        # Group the selectors and the dataset selector into a container.
        self._content = Container(self._widgets)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def disable(self):
        """Disable the widget in the UI."""
        for widget in self._widgets:
            widget.disable()

    def enable(self) -> None:
        """Enable the widget in the UI."""
        for widget in self._widgets:
            widget.enable()

    @property
    def team_id(self) -> int:
        """The ID of the team selected in the widget.

        :return: The ID of the team.
        :rtype: int
        """
        return self._team_id

    @team_id.setter
    def team_id(self, team_id: int) -> None:
        """Set the team ID to read workspaces from.

        :param team_id: The ID of the team.
        :type team_id: int
        """
        if not self._compact:
            self._select_team.set_value(team_id)
            self._select_workspace.set(self._get_select_items(team_id=team_id))
        self._team_id = team_id

    def get_selected_team_id(self) -> int:
        """Get the ID of the selected team.

        :return: The ID of the selected team.
        :rtype: int
        """
        return self.team_id

    def set_team_id(self, team_id: int) -> None:
        """Set the team ID to read workspaces from.

        :param team_id: The ID of the team.
        :type team_id: int
        """
        self.team_id = team_id

    @property
    def workspace_id(self) -> int:
        """The ID of the workspace selected in the widget.

        :return: The ID of the workspace.
        :rtype: int
        """
        return self._workspace_id

    @workspace_id.setter
    def workspace_id(self, workspace_id: int) -> None:
        """Set the workspace ID to read projects from.

        :param workspace_id: The ID of the workspace.
        :type workspace_id: int
        """
        if not self._compact:
            self._select_workspace.set_value(workspace_id)
            self._select_project.set(self._get_select_items(workspace_id=workspace_id))
        self._workspace_id = workspace_id

    def get_selected_workspace_id(self) -> int:
        """Get the ID of the selected workspace.

        :return: The ID of the selected workspace.
        :rtype: int
        """
        return self.workspace_id

    def set_workspace_id(self, workspace_id: int) -> None:
        """Set the workspace ID to read projects from.

        :param workspace_id: The ID of the workspace.
        :type workspace_id: int
        """
        self.workspace_id = workspace_id

    @property
    def project_id(self) -> Optional[int]:
        """The ID of the project selected in the widget.

        :return: The ID of the project.
        :rtype: Optional[int]
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id: Optional[int]) -> None:
        """Set the project ID to read datasets from.

        :param project_id: The ID of the project.
        :type project_id: int
        """
        if not self._compact:
            self._select_project.set_value(project_id)
        self._project_id = project_id
        self._select_dataset.set_items(self._read_datasets(project_id))

    def get_selected_project_id(self) -> Optional[int]:
        """Get the ID of the selected project.

        :return: The ID of the selected project.
        :rtype: Optional[int]
        """
        return self.project_id

    def set_dataset_id(self, dataset_id: int) -> None:
        """Set the ID of the dataset to be selected by default.

        :param id: The ID of the dataset.
        :type id: int
        """
        self._select_dataset.set_selected_by_id(dataset_id)

    def set_dataset_ids(self, dataset_ids: List[int]) -> None:
        """Set the IDs of the datasets to be selected by default.

        :raise ValueError: If multiselect is disabled.
        :param ids: The IDs of the datasets.
        :type ids: List[int]
        """
        if not self._multiselect:
            raise ValueError("This method can only be called when multiselect is enabled.")
        self._select_dataset.set_selected_by_id(dataset_ids)

    def value_changed(self, func: Callable) -> Callable:
        """Decorator to set the callback function for the value changed event.

        :param func: The callback function.
        :type func: Callable
        :return: The callback function.
        :rtype: Callable
        """

        @self._select_dataset.value_changed
        def _click(items: Union[List[TreeSelect.Item], TreeSelect.Item]):
            if isinstance(items, list):
                res = [item.id for item in items]
            else:
                res = items.id

            func(res)

        return _click

    def _create_select_all_datasets_checkbox(self, select_all_datasets: bool) -> None:
        """Create the checkbox to select all datasets.

        :param select_all_datasets: Whether all datasets should be selected by default.
        :type select_all_datasets: bool
        """
        if not self._multiselect:
            # We'll only create the checkbox if multiselect is enabled.
            return
        select_all_datasets_checkbox = Checkbox("Select all datasets")

        @select_all_datasets_checkbox.value_changed
        def select_all_datasets_checkbox_handler(checked: bool) -> None:
            """Handler function for the event when the checkbox value changes.

            :param checked: The value of the checkbox.
            :type checked: bool
            """
            if self._project_id is None:
                return

            if checked:
                self._select_dataset.select_all()
                self._select_dataset.hide()
            else:
                self._select_dataset.clear_selected()
                self._select_dataset.show()

        if select_all_datasets:
            self._select_dataset.hide()
            select_all_datasets_checkbox.check()

        self._widgets.append(select_all_datasets_checkbox)
        self._select_all_datasets_checkbox = select_all_datasets_checkbox

    def _create_dataset_selector(
        self, flat: bool, always_open: bool, select_all_datasets: bool
    ) -> None:
        """Create the dataset selector.

        :param flat: Whether the dataset selector should be flat.
        :type flat: bool
        :param always_open: Whether the dataset selector should always be open.
        :type always_open: bool
        :param select_all_datasets: Whether all datasets should be selected by default.
        :type select_all_datasets: bool
        """
        self._select_dataset = TreeSelect(
            items=self._read_datasets(self._project_id),
            multiple_select=self._multiselect,
            flat=flat,
            always_open=always_open,
            append_to_body=self._append_to_body,
            width=self._width,
            placeholder="Select dataset",
        )
        if self._dataset_id is not None:
            self._select_dataset.set_selected_by_id(self._dataset_id)
        if select_all_datasets:
            self._select_dataset.select_all()

        # Adding the dataset selector to the list of widgets to be added to the container.
        self._widgets.append(self._select_dataset)

    def _create_selectors(self, team_is_selectable: bool, workspace_is_selectable: bool):
        """Create the team, workspace, and project selectors.

        :param team_is_selectable: Whether the team selector should be selectable.
        :type team_is_selectable: bool
        :param workspace_is_selectable: Whether the workspace selector should be selectable.
        :type workspace_is_selectable: bool
        """

        def team_selector_handler(team_id: int):
            """Handler function for the event when the team selector value changes.

            :param team_id: The ID of the selected team.
            :type team_id: int
            """
            self._select_workspace.set(items=self._get_select_items(team_id=team_id))
            self._team_id = team_id

        def workspace_selector_handler(workspace_id: int):
            """Handler function for the event when the workspace selector value changes.

            :param workspace_id: The ID of the selected workspace.
            :type workspace_id: int
            """
            self._select_project.set(items=self._get_select_items(workspace_id=workspace_id))
            self._workspace_id = workspace_id

        def project_selector_handler(project_id: int):
            """Handler function for the event when the project selector value changes.

            :param project_id: The ID of the selected project.
            :type project_id: int
            """
            self._select_dataset.set_items(self._read_datasets(project_id))
            self._project_id = project_id

            if (
                self._select_all_datasets_checkbox is not None
                and self._select_all_datasets_checkbox.is_checked()
            ):
                self._select_dataset.select_all()
                self._select_dataset.hide()

        self._select_team = Select(
            items=self._get_select_items(),
            placeholder="Select team",
            filterable=True,
            width_px=self._width,
        )
        self._select_team.set_value(self._team_id)
        if not team_is_selectable:
            self._select_team.disable()

        self._select_workspace = Select(
            items=self._get_select_items(team_id=self._team_id),
            placeholder="Select workspace",
            filterable=True,
            width_px=self._width,
        )
        self._select_workspace.set_value(self._workspace_id)
        if not workspace_is_selectable:
            self._select_workspace.disable()

        self._select_project = Select(
            items=self._get_select_items(workspace_id=self._workspace_id),
            placeholder="Select project",
            filterable=True,
            width_px=self._width,
        )
        self._select_project.set_value(self._project_id)

        # Register the event handlers.
        self._select_team.value_changed(team_selector_handler)
        self._select_workspace.value_changed(workspace_selector_handler)
        self._select_project.value_changed(project_selector_handler)

        # Adding widgets to the list, so they can be added to the container.
        self._widgets.extend([self._select_team, self._select_workspace, self._select_project])

    def _get_select_items(self, **kwargs) -> List[Select.Item]:
        """Get the list of items for the team, workspace, and project selectors.
        Possible keyword arguments are 'team_id' and 'workspace_id'.

        :return: The list of items.
        :rtype: List[Select.Item]
        """
        if not kwargs:
            items = self._api.team.get_list()
        elif "team_id" in kwargs:
            items = self._api.workspace.get_list(kwargs["team_id"])
        elif "workspace_id" in kwargs:
            projects_list = self._api.project.get_list(kwargs["workspace_id"])
            if self._project_types is not None:
                items = [
                    project for project in projects_list if project.type in self._project_types
                ]  # TODO: Filter project from API, not from here.
            else:
                items = projects_list

        return [Select.Item(value=item.id, label=item.name) for item in items]

    def get_json_data(self) -> Dict:
        """Get the JSON data of the widget.

        :return: The JSON data.
        :rtype: Dict
        """
        return {}

    def get_json_state(self) -> Dict:
        """Get the JSON state of the widget.

        :return: The JSON state.
        :rtype: Dict
        """
        return {}

    def _read_datasets(self, project_id: Optional[int]) -> Optional[List[TreeSelect.Item]]:
        """Get the lisf of TreeSelect.Item objects representing the dataset hierarchy.

        :param project_id: The ID of the project.
        :type project_id: Optional[int]
        :return: The list of TreeSelect.Item objects.
        :rtype: Optional[List[TreeSelect.Item]]
        """
        if not project_id:
            return None
        dataset_tree = self._api.dataset.get_tree(project_id)

        def convert_tree_to_list(node, parent_id: Optional[int] = None):
            """
            Recursively converts a tree of DatasetInfo objects into a list of
                SelectDatasetTree.Item objects.

            :param node: The current node in the tree (a tuple of DatasetInfo and its children).
            :param parent_id: The ID of the parent dataset, if any.
            :return: A list of SelectDatasetTree.Item objects representing the tree.
            """
            result = []
            for dataset_info, children in node.items():
                item = TreeSelect.Item(
                    id=dataset_info.id,
                    label=dataset_info.name,
                    children=convert_tree_to_list(children, parent_id=dataset_info.id),
                )

                result.append(item)

            return result

        return convert_tree_to_list(dataset_tree)

    def set_project_id(self, project_id: Optional[int]) -> None:
        """Set the project ID to read datasets from.

        :param project_id: The ID of the project.
        :type project_id: int
        """
        self.project_id = project_id

    def _get_selected(self) -> Optional[Union[List[int], int]]:
        """Get the ID of the selected dataset(s).

        :return: The ID of the selected dataset(s).
        :rtype: Optional[Union[List[int], int]]
        """
        selected = self._select_dataset.get_selected()
        if not selected:
            return None

        if isinstance(selected, list):
            return [item.id for item in selected]
        else:
            return selected.id

    def get_selected_ids(self) -> Optional[List[int]]:
        """Get the IDs of the selected datasets.

        raise ValueError if multiselect is disabled.
        return: The IDs of the selected datasets.
        rtype: Optional[List[int]]
        """
        if not self._multiselect:
            raise ValueError("This method can only be called when multiselect is enabled.")
        return self._get_selected()

    def get_selected_id(self) -> Optional[int]:
        """Get the ID of the selected dataset.

        raise ValueError if multiselect is enabled.
        return: The ID of the selected dataset.
        rtype: Optional[int]
        """
        if self._multiselect:
            raise ValueError("This method can only be called when multiselect is disabled.")
        return self._get_selected()

    def is_all_selected(self) -> bool:
        """Check if all datasets are selected.

        return: True if all datasets are selected, False otherwise.
        rtype: bool
        """
        return self._select_dataset.is_all_selected()

    def select_all(self) -> None:
        """Select all datasets."""
        self._select_dataset.select_all()



from supervisely.nn.inference.gui.gui import BaseInferenceGUI, InferenceGUI
from supervisely.nn.inference.gui.serving_gui import ServingGUI
from supervisely.nn.inference.gui.serving_gui_template import ServingGUITemplate



import os
from pathlib import Path

from supervisely import ProjectMeta, generate_free_name, logger
from supervisely._utils import batched, is_development
from supervisely.annotation.obj_class import ObjClass
from supervisely.api.api import Api
from supervisely.convert.base_converter import AvailableVolumeConverters
from supervisely.convert.volume.nii import nii_volume_helper as helper
from supervisely.convert.volume.volume_converter import VolumeConverter
from supervisely.geometry.mask_3d import Mask3D
from supervisely.io.fs import (
    get_file_ext,
    get_file_name,
    get_file_name_with_ext,
    list_files,
)
from supervisely.task.progress import tqdm_sly
from supervisely.volume.volume import is_nifti_file, read_nrrd_serie_volume_np
from supervisely.volume_annotation.volume_annotation import VolumeAnnotation
from supervisely.volume_annotation.volume_object import VolumeObject


class NiiConverter(VolumeConverter):
    """
    Convert NIfTI 3D volume file to Supervisely format.
    Supports .nii and .nii.gz files.

    The NIfTI file should be structured as follows:
    - <volume_name>.nii
    - <volume_name>/
        - <cls_name_1>.nii
        - <cls_name_2>.nii
        - ...
    - ...

    where   <volume_name> is the name of the volume
            If the volume has annotations, they should be in the corresponding directory
                with the same name as the volume (without extension)
            <cls_name> is the name of the annotation class
                <cls_name>.nii:
                    - represent objects of the single class
                    - should be unique for the current volume (e.g. tumor.nii.gz, lung.nii.gz)
                    - can contain multiple objects of the class (each object should be represented by a different value in the mask)

    Example:
    📂 .
    ├── 📂 CTChest
    │   ├── 🩻 lung.nii.gz
    │   └── 🩻 tumor.nii.gz
    ├── 🩻 CTChest.nii.gz
    └── 🩻 Spine.nii.gz
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._supports_links = True
        self._meta_changed = False

    def __str__(self) -> str:
        return AvailableVolumeConverters.NII

    def validate_format(self) -> bool:
        # create Items
        converted_dir_name = "converted"
        # nrrds_dict = {}
        nifti_dict = {}
        nifti_dirs = {}

        planes_detected = {p: False for p in helper.PlanePrefix.values()}

        for root, _, files in os.walk(self._input_data):
            dir_name = os.path.basename(root)
            nifti_dirs[dir_name] = root
            if converted_dir_name in root:
                continue
            for file in files:
                path = os.path.join(root, file)
                if is_nifti_file(path):  # is nifti
                    name = get_file_name(path)
                    if name.endswith(".nii"):
                        name = get_file_name(name)
                    nifti_dict[name] = path
                    for plane in planes_detected.keys():
                        if plane in name:
                            planes_detected[plane] = True

        if any(planes_detected.values()):
            return False

        self._items = []
        skip_files = []
        for name, nrrd_path in nifti_dict.items():
            if name in nifti_dirs:
                item = self.Item(item_path=nrrd_path)
                ann_dir = nifti_dirs[name]
                item.ann_data = list_files(ann_dir, [".nii", ".nii.gz", ".gz"], None, True)
                self._items.append(item)
                skip_files.extend(item.ann_data)
                skip_files.append(nrrd_path)

        for name, nrrd_path in nifti_dict.items():
            if nrrd_path in skip_files:
                continue
            item = self.Item(item_path=nrrd_path)
            self._items.append(item)

        self._meta = ProjectMeta()
        return self.items_count > 0

    def to_supervisely(
        self,
        item: VolumeConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> VolumeAnnotation:
        """Convert to Supervisely format."""

        try:
            objs = []
            spatial_figures = []
            for ann_path in item.ann_data:
                ann_name = get_file_name(ann_path)
                if ann_name.endswith(".nii"):
                    ann_name = get_file_name(ann_name)

                ann_name = renamed_classes.get(ann_name, ann_name)
                for mask, _ in helper.get_annotation_from_nii(ann_path):
                    obj_class = meta.get_obj_class(ann_name)
                    if obj_class is None:
                        obj_class = ObjClass(ann_name, Mask3D)
                        meta = meta.add_obj_class(obj_class)
                        self._meta_changed = True
                        self._meta = meta
                    obj = VolumeObject(obj_class, mask_3d=mask)
                    spatial_figures.append(obj.figure)
                    objs.append(obj)
            return VolumeAnnotation(item.volume_meta, objects=objs, spatial_figures=spatial_figures)
        except Exception as e:
            logger.warning(f"Failed to convert {item.path} to Supervisely format: {e}")
            return item.create_empty_annotation()

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 1,
        log_progress=True,
    ):
        """Upload converted data to Supervisely"""

        meta, renamed_classes, _ = self.merge_metas_with_conflicts(api, dataset_id)

        existing_names = set([vol.name for vol in api.volume.get_list(dataset_id)])

        if log_progress:
            progress, progress_cb = self.get_progress(
                self.items_count, "Converting and uploading volumes..."
            )
        else:
            progress_cb = None

        converted_dir_name = "converted"
        converted_dir = os.path.join(self._input_data, converted_dir_name)

        for batch in batched(self._items, batch_size=batch_size):
            item_names = []
            item_paths = []

            for item in batch:
                # if self._upload_as_links:
                #     remote_path = self.remote_files_map.get(item.path)
                #     if remote_path is not None:
                #         item.custom_data = {"remote_path": remote_path}

                item.path = helper.nifti_to_nrrd(item.path, converted_dir)
                ext = get_file_ext(item.path)
                if ext.lower() != ext:
                    new_volume_path = Path(item.path).with_suffix(ext.lower()).as_posix()
                    os.rename(item.path, new_volume_path)
                    item.path = new_volume_path
                item.name = get_file_name_with_ext(item.path)
                item.name = generate_free_name(
                    existing_names, item.name, with_ext=True, extend_used_names=True
                )
                item_names.append(item.name)
                item_paths.append(item.path)

                # upload volume
                volume_np, volume_meta = read_nrrd_serie_volume_np(item.path)
                progress_nrrd = tqdm_sly(
                    desc=f"Uploading volume '{item.name}'",
                    total=sum(volume_np.shape),
                    leave=True if progress_cb is None else False,
                    position=1,
                )
                # if item.custom_data is not None:
                #     volume_meta.update(item.custom_data)
                api.volume.upload_np(dataset_id, item.name, volume_np, volume_meta, progress_nrrd)
                info = api.volume.get_info_by_name(dataset_id, item.name)
                item.volume_meta = info.meta

                # create and upload annotation
                if item.ann_data is not None:
                    ann = self.to_supervisely(item, meta, renamed_classes, None)

                    if self._meta_changed:
                        meta, renamed_classes, _ = self.merge_metas_with_conflicts(api, dataset_id)
                        self._meta_changed = False
                    api.volume.annotation.append(info.id, ann)

            if log_progress:
                progress_cb(len(batch))

        if log_progress:
            if is_development():
                progress.close()
        logger.info(f"Dataset ID:{dataset_id} has been successfully uploaded.")



"""
GUI module for training application.

This module provides the `TrainGUI` class that handles the graphical user interface (GUI) for managing
training workflows in Supervisely.
"""

from os import environ
from typing import Union

import supervisely.io.env as sly_env
import supervisely.io.json as sly_json
from supervisely import Api, ProjectMeta
from supervisely._utils import is_production
from supervisely.app.widgets import Stepper, Widget
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.rectangle import Rectangle
from supervisely.nn.task_type import TaskType
from supervisely.nn.training.gui.classes_selector import ClassesSelector
from supervisely.nn.training.gui.hyperparameters_selector import HyperparametersSelector
from supervisely.nn.training.gui.input_selector import InputSelector
from supervisely.nn.training.gui.model_selector import ModelSelector
from supervisely.nn.training.gui.train_val_splits_selector import TrainValSplitsSelector
from supervisely.nn.training.gui.training_artifacts import TrainingArtifacts
from supervisely.nn.training.gui.training_logs import TrainingLogs
from supervisely.nn.training.gui.training_process import TrainingProcess
from supervisely.nn.training.gui.utils import set_stepper_step, wrap_button_click
from supervisely.nn.utils import ModelSource, RuntimeType


class TrainGUI:
    """
    A class representing the GUI for training workflows.

    This class sets up and manages GUI components such as project selection,
    train/validation split selection, model selection, hyperparameters selection,
    and the training process.

    :param framework_name: Name of the ML framework being used.
    :type framework_name: str
    :param models: List of available models.
    :type models: list
    :param hyperparameters: Hyperparameters for training.
    :type hyperparameters: dict
    :param app_options: Application options for customization.
    :type app_options: dict, optional
    """

    def __init__(
        self,
        framework_name: str,
        models: list,
        hyperparameters: dict,
        app_options: dict = None,
    ):
        self._api = Api.from_env()
        if is_production():
            self.task_id = sly_env.task_id()
        else:
            self.task_id = sly_env.task_id(raise_not_found=False)
            if self.task_id is None:
                self.task_id = "debug-session"

        self.framework_name = framework_name
        self.models = models
        self.hyperparameters = hyperparameters
        self.app_options = app_options
        self.collapsable = self.app_options.get("collapsable", False)
        self.need_convert_shapes_for_bm = False

        self.team_id = sly_env.team_id(raise_not_found=False)
        self.workspace_id = sly_env.workspace_id(raise_not_found=False)
        self.project_id = sly_env.project_id()
        self.project_info = self._api.project.get_info_by_id(self.project_id)
        self.project_meta = ProjectMeta.from_json(self._api.project.get_meta(self.project_id))

        if self.workspace_id is None:
            self.workspace_id = self.project_info.workspace_id
            environ["WORKSPACE_ID"] = str(self.workspace_id)
        if self.team_id is None:
            self.team_id = self.project_info.team_id
            environ["TEAM_ID"] = str(self.team_id)

        # 1. Project selection + Train/val split
        self.input_selector = InputSelector(self.project_info, self.app_options)
        # 2. Select train val splits
        self.train_val_splits_selector = TrainValSplitsSelector(
            self._api, self.project_id, self.app_options
        )
        # 3. Select classes
        self.classes_selector = ClassesSelector(self.project_id, [], self.app_options)
        # 4. Model selection
        self.model_selector = ModelSelector(
            self._api, self.framework_name, self.models, self.app_options
        )
        # 5. Training parameters (yaml), scheduler preview
        self.hyperparameters_selector = HyperparametersSelector(
            self.hyperparameters, self.app_options
        )
        # 6. Start Train
        self.training_process = TrainingProcess(self.app_options)

        # 7. Training logs
        self.training_logs = TrainingLogs(self.app_options)

        # 8. Training Artifacts
        self.training_artifacts = TrainingArtifacts(self._api, self.app_options)

        # Stepper layout
        self.steps = [
            self.input_selector.card,
            self.train_val_splits_selector.card,
            self.classes_selector.card,
            self.model_selector.card,
            self.hyperparameters_selector.card,
            self.training_process.card,
            self.training_logs.card,
            self.training_artifacts.card,
        ]
        self.stepper = Stepper(
            widgets=self.steps,
        )
        # ------------------------------------------------- #

        # Button utils
        def update_classes_table():
            pass

        def disable_hyperparams_editor():
            if self.hyperparameters_selector.editor.readonly:
                self.hyperparameters_selector.editor.readonly = False
            else:
                self.hyperparameters_selector.editor.readonly = True

        def set_experiment_name():
            model_name = self.model_selector.get_model_name()
            if model_name is None:
                experiment_name = "Enter experiment name"
            else:
                experiment_name = f"{self.task_id}_{self.project_info.name}_{model_name}"

            if experiment_name == self.training_process.get_experiment_name():
                return
            self.training_process.set_experiment_name(experiment_name)

        def need_convert_class_shapes() -> bool:
            if self.hyperparameters_selector.run_model_benchmark_checkbox is not None:
                if not self.hyperparameters_selector.run_model_benchmark_checkbox.is_checked():
                    self.hyperparameters_selector.model_benchmark_auto_convert_warning.hide()
                    self.need_convert_shapes_for_bm = False
                else:
                    task_type = self.model_selector.get_selected_task_type()

                    def _need_convert(shape):
                        if task_type == TaskType.OBJECT_DETECTION:
                            return shape != Rectangle.geometry_name()
                        elif task_type in [
                            TaskType.INSTANCE_SEGMENTATION,
                            TaskType.SEMANTIC_SEGMENTATION,
                        ]:
                            return shape == Polygon.geometry_name()
                        return

                    data = self.classes_selector.classes_table._table_data
                    selected_classes = set(
                        self.classes_selector.classes_table.get_selected_classes()
                    )
                    empty = set(
                        r[0]["data"] for r in data if r[2]["data"] == 0 and r[3]["data"] == 0
                    )
                    need_convert = set(r[0]["data"] for r in data if _need_convert(r[1]["data"]))

                    if need_convert.intersection(selected_classes - empty):
                        self.hyperparameters_selector.model_benchmark_auto_convert_warning.show()
                        self.need_convert_shapes_for_bm = True
                    else:
                        self.hyperparameters_selector.model_benchmark_auto_convert_warning.hide()
                        self.need_convert_shapes_for_bm = False
            else:
                self.need_convert_shapes_for_bm = False

        def validate_class_shape_for_model_task():
            task_type = self.model_selector.get_selected_task_type()
            classes = self.classes_selector.get_selected_classes()

            required_geometries = {
                TaskType.INSTANCE_SEGMENTATION: {Polygon, Bitmap},
                TaskType.SEMANTIC_SEGMENTATION: {Polygon, Bitmap},
                TaskType.POSE_ESTIMATION: {GraphNodes},
            }
            task_specific_texts = {
                TaskType.INSTANCE_SEGMENTATION: "Only polygon and bitmap shapes are supported for segmentation task",
                TaskType.SEMANTIC_SEGMENTATION: "Only polygon and bitmap shapes are supported for segmentation task",
                TaskType.POSE_ESTIMATION: "Only keypoint (graph) shape is supported for pose estimation task",
            }

            if task_type not in required_geometries:
                return

            wrong_shape_classes = [
                class_name
                for class_name in classes
                if self.project_meta.get_obj_class(class_name).geometry_type
                not in required_geometries[task_type]
            ]

            if wrong_shape_classes:
                specific_text = task_specific_texts[task_type]
                message_text = f"Model task type is {task_type}. {specific_text}. Selected classes have wrong shapes for the model task: {', '.join(wrong_shape_classes)}"
                self.model_selector.validator_text.set(
                    text=message_text,
                    status="warning",
                )

        # ------------------------------------------------- #

        # Wrappers
        self.training_process_cb = wrap_button_click(
            button=self.hyperparameters_selector.button,
            cards_to_unlock=[self.training_logs.card],
            widgets_to_disable=self.training_process.widgets_to_disable,
            callback=None,
            validation_text=self.training_process.validator_text,
            validation_func=self.training_process.validate_step,
        )

        self.hyperparameters_selector_cb = wrap_button_click(
            button=self.hyperparameters_selector.button,
            cards_to_unlock=[self.training_process.card],
            widgets_to_disable=self.hyperparameters_selector.widgets_to_disable,
            callback=self.training_process_cb,
            validation_text=self.hyperparameters_selector.validator_text,
            validation_func=self.hyperparameters_selector.validate_step,
            on_select_click=[disable_hyperparams_editor],
            on_reselect_click=[disable_hyperparams_editor],
            collapse_card=(self.hyperparameters_selector.card, self.collapsable),
        )

        self.model_selector_cb = wrap_button_click(
            button=self.model_selector.button,
            cards_to_unlock=[self.hyperparameters_selector.card],
            widgets_to_disable=self.model_selector.widgets_to_disable,
            callback=self.hyperparameters_selector_cb,
            validation_text=self.model_selector.validator_text,
            validation_func=self.model_selector.validate_step,
            on_select_click=[
                set_experiment_name,
                need_convert_class_shapes,
                validate_class_shape_for_model_task,
            ],
            collapse_card=(self.model_selector.card, self.collapsable),
        )

        self.classes_selector_cb = wrap_button_click(
            button=self.classes_selector.button,
            cards_to_unlock=[self.model_selector.card],
            widgets_to_disable=self.classes_selector.widgets_to_disable,
            callback=self.model_selector_cb,
            validation_text=self.classes_selector.validator_text,
            validation_func=self.classes_selector.validate_step,
            collapse_card=(self.classes_selector.card, self.collapsable),
        )

        self.train_val_splits_selector_cb = wrap_button_click(
            button=self.train_val_splits_selector.button,
            cards_to_unlock=[self.classes_selector.card],
            widgets_to_disable=self.train_val_splits_selector.widgets_to_disable,
            callback=self.classes_selector_cb,
            validation_text=self.train_val_splits_selector.validator_text,
            validation_func=self.train_val_splits_selector.validate_step,
            collapse_card=(self.train_val_splits_selector.card, self.collapsable),
        )

        self.input_selector_cb = wrap_button_click(
            button=self.input_selector.button,
            cards_to_unlock=[self.train_val_splits_selector.card],
            widgets_to_disable=self.input_selector.widgets_to_disable,
            callback=self.train_val_splits_selector_cb,
            validation_text=self.input_selector.validator_text,
            validation_func=self.input_selector.validate_step,
            on_select_click=[update_classes_table],
            collapse_card=(self.input_selector.card, self.collapsable),
        )
        # ------------------------------------------------- #

        # Main Buttons

        # Define outside. Used by user in app
        # @self.training_process.start_button.click
        # def start_training():
        #     pass

        # @self.training_process.stop_button.click
        # def stop_training():
        #     pass

        # ------------------------------------------------- #

        # Select Buttons
        @self.hyperparameters_selector.button.click
        def select_hyperparameters():
            self.hyperparameters_selector_cb()
            set_stepper_step(
                self.stepper,
                self.hyperparameters_selector.button,
                next_pos=6,
            )

        @self.model_selector.button.click
        def select_model():
            self.model_selector_cb()
            set_stepper_step(
                self.stepper,
                self.model_selector.button,
                next_pos=5,
            )

        @self.classes_selector.button.click
        def select_classes():
            self.classes_selector_cb()
            set_stepper_step(
                self.stepper,
                self.classes_selector.button,
                next_pos=4,
            )

        @self.train_val_splits_selector.button.click
        def select_train_val_splits():
            self.train_val_splits_selector_cb()
            set_stepper_step(
                self.stepper,
                self.train_val_splits_selector.button,
                next_pos=3,
            )

        @self.input_selector.button.click
        def select_input():
            self.input_selector_cb()
            set_stepper_step(
                self.stepper,
                self.input_selector.button,
                next_pos=2,
            )

        # ------------------------------------------------- #

        # Other Buttons
        if self.app_options.get("show_logs_in_gui", False):

            @self.training_logs.logs_button.click
            def show_logs():
                self.training_logs.toggle_logs()

        # Other handlers
        if self.hyperparameters_selector.run_model_benchmark_checkbox is not None:

            @self.hyperparameters_selector.run_model_benchmark_checkbox.value_changed
            def show_mb_speedtest(is_checked: bool):
                self.hyperparameters_selector.toggle_mb_speedtest(is_checked)
                need_convert_class_shapes()

        # ------------------------------------------------- #

        self.layout: Widget = self.stepper

    def set_next_step(self):
        current_step = self.stepper.get_active_step()
        self.stepper.set_active_step(current_step + 1)

    def set_previous_step(self):
        current_step = self.stepper.get_active_step()
        self.stepper.set_active_step(current_step - 1)

    def set_first_step(self):
        self.stepper.set_active_step(1)

    def set_last_step(self):
        self.stepper.set_active_step(len(self.steps))

    def enable_select_buttons(self):
        """
        Makes all select buttons in the GUI available for interaction.
        """
        self.input_selector.button.enable()
        self.train_val_splits_selector.button.enable()
        self.classes_selector.button.enable()
        self.model_selector.button.enable()
        self.hyperparameters_selector.button.enable()

    def disable_select_buttons(self):
        """
        Makes all select buttons in the GUI unavailable for interaction.
        """
        self.input_selector.button.disable()
        self.train_val_splits_selector.button.disable()
        self.classes_selector.button.disable()
        self.model_selector.button.disable()
        self.hyperparameters_selector.button.disable()

    # Set GUI from config
    def validate_app_state(self, app_state: dict) -> dict:
        """
        Validate the app state dictionary.

        :param app_state: The app state dictionary.
        :type app_state: dict
        """
        if not isinstance(app_state, dict):
            raise ValueError("app_state must be a dictionary")

        required_keys = {
            "train_val_split": ["method"],
            "classes": list,
            "model": ["source"],
            "hyperparameters": (dict, str),  # Allowing dict or str for hyperparameters
        }

        for key, subkeys_or_type in required_keys.items():
            if key not in app_state:
                raise KeyError(f"Missing required key in app_state: {key}")

            if isinstance(subkeys_or_type, list):
                for subkey in subkeys_or_type:
                    if subkey not in app_state[key]:
                        raise KeyError(f"Missing required key in app_state['{key}']: {subkey}")
            elif not isinstance(app_state[key], subkeys_or_type):
                valid_types = (
                    " or ".join([t.__name__ for t in subkeys_or_type])
                    if isinstance(subkeys_or_type, tuple)
                    else subkeys_or_type.__name__
                )
                raise ValueError(f"app_state['{key}'] must be of type {valid_types}")

        model = app_state["model"]
        if model["source"] == "Pretrained models":
            if "model_name" not in model:
                raise KeyError("Missing required key in app_state['model']: model_name")
        elif model["source"] == "Custom models":
            custom_keys = ["task_id", "checkpoint"]
            for key in custom_keys:
                if key not in model:
                    raise KeyError(f"Missing required key in app_state['model']: {key}")

        options = app_state.setdefault(
            "options",
            {
                "model_benchmark": {
                    "enable": True,
                    "speed_test": True,
                },
                "cache_project": True,
            },
        )

        if not isinstance(options, dict):
            raise ValueError("app_state['options'] must be a dictionary")

        model_benchmark = options.setdefault(
            "model_benchmark", {"enable": True, "speed_test": True}
        )
        if not isinstance(model_benchmark, dict):
            raise ValueError("app_state['options']['model_benchmark'] must be a dictionary")
        model_benchmark.setdefault("enable", True)
        model_benchmark.setdefault("speed_test", True)

        if not isinstance(options.get("cache_project"), bool):
            raise ValueError("app_state['options']['cache_project'] must be a boolean")

        # Check train val splits
        train_val_splits_settings = app_state.get("train_val_split")
        if train_val_splits_settings.get("method") == "datasets":
            dataset_ids = []
            for parents, dataset in self._api.dataset.tree(self.project_id):
                dataset_ids.append(dataset.id)

            train_datasets = train_val_splits_settings.get("train_datasets", [])
            val_datasets = train_val_splits_settings.get("val_datasets", [])

            missing_datasets_ids = []
            for ds_id in train_datasets + val_datasets:
                if ds_id not in dataset_ids:
                    missing_datasets_ids.append(ds_id)

            if len(missing_datasets_ids) > 0:
                missing_datasets_text = ", ".join([str(ds_id) for ds_id in missing_datasets_ids])
                raise ValueError(
                    f"Datasets with ids: {missing_datasets_text} not found in the project"
                )
        elif train_val_splits_settings.get("method") == "tags":
            train_tag = train_val_splits_settings.get("train_tag")
            val_tag = train_val_splits_settings.get("val_tag")
            if not train_tag or not val_tag:
                raise ValueError("train_tag and val_tag must be specified in tags split method")
        elif train_val_splits_settings.get("method") == "random":
            split = train_val_splits_settings.get("split")
            percent = train_val_splits_settings.get("percent")
            if split not in ["train", "val"]:
                raise ValueError("split must be 'train' or 'val'")
            if not isinstance(percent, int) or not 0 < percent < 100:
                raise ValueError("percent must be an integer in range 1 to 99")
        return app_state

    def load_from_app_state(self, app_state: Union[str, dict]) -> None:
        """
        Load the GUI state from app state dictionary.

        :param app_state: The state dictionary.
        :type app_state: dict

        app_state example:

            app_state = {
                "train_val_split": {
                    "method": "random",
                    "split": "train",
                    "percent": 90
                },
                "classes": ["apple"],
                "model": {
                    "source": "Pretrained models",
                    "model_name": "rtdetr_r50vd_coco_objects365"
                },
                "hyperparameters": hyperparameters, # yaml string
                "options": {
                    "model_benchmark": {
                        "enable": True,
                        "speed_test": True
                    },
                    "cache_project": True,
                "export": {
                    "enable": True,
                    "ONNXRuntime": True,
                    "TensorRT": True
                    },
                }
            }
        """
        if isinstance(app_state, str):
            app_state = sly_json.load_json_file(app_state)

        app_state = self.validate_app_state(app_state)

        options = app_state.get("options", {})
        input_settings = app_state.get("input")
        train_val_splits_settings = app_state["train_val_split"]
        classes_settings = app_state["classes"]
        model_settings = app_state["model"]
        hyperparameters_settings = app_state["hyperparameters"]

        self._init_input(input_settings, options)
        self._init_classes(classes_settings)
        self._init_train_val_splits(train_val_splits_settings)
        self._init_model(model_settings)
        self._init_hyperparameters(hyperparameters_settings, options)

    def _init_input(self, input_settings: Union[dict, None], options: dict) -> None:
        """
        Initialize the input selector with the given settings.

        :param input_settings: The input settings.
        :type input_settings: dict
        :param options: The application options.
        :type options: dict
        """
        # Set Input
        self.input_selector.set_cache(options.get("cache_project", True))
        self.input_selector_cb()
        # ----------------------------------------- #

    def _init_train_val_splits(self, train_val_splits_settings: dict) -> None:
        """
        Initialize the train/val splits selector with the given settings.

        :param train_val_splits_settings: The train/val splits settings.
        :type train_val_splits_settings: dict
        """
        split_method = train_val_splits_settings["method"]
        if split_method == "random":
            split = train_val_splits_settings["split"]
            percent = train_val_splits_settings["percent"]
            self.train_val_splits_selector.train_val_splits.set_random_splits(split, percent)
        elif split_method == "tags":
            train_tag = train_val_splits_settings["train_tag"]
            val_tag = train_val_splits_settings["val_tag"]
            untagged_action = train_val_splits_settings["untagged_action"]
            self.train_val_splits_selector.train_val_splits.set_tags_splits(
                train_tag, val_tag, untagged_action
            )
        elif split_method == "datasets":
            train_datasets = train_val_splits_settings["train_datasets"]
            val_datasets = train_val_splits_settings["val_datasets"]
            self.train_val_splits_selector.train_val_splits.set_datasets_splits(
                train_datasets, val_datasets
            )
        self.train_val_splits_selector_cb()

    def _init_classes(self, classes_settings: list) -> None:
        """
        Initialize the classes selector with the given settings.

        :param classes_settings: The classes settings.
        :type classes_settings: list
        """
        # Set Classes
        self.classes_selector.set_classes(classes_settings)
        self.classes_selector_cb()
        # ----------------------------------------- #

    def _init_model(self, model_settings: dict) -> None:
        """
        Initialize the model selector with the given settings.

        :param model_settings: The model settings.
        :type model_settings: dict
        """

        # Pretrained
        if model_settings["source"] == ModelSource.PRETRAINED:
            self.model_selector.model_source_tabs.set_active_tab(ModelSource.PRETRAINED)
            self.model_selector.pretrained_models_table.set_by_model_name(
                model_settings["model_name"]
            )

        # Custom
        elif model_settings["source"] == ModelSource.CUSTOM:
            self.model_selector.model_source_tabs.set_active_tab(ModelSource.CUSTOM)
            self.model_selector.experiment_selector.set_by_task_id(model_settings["task_id"])
            active_row = self.model_selector.experiment_selector.get_selected_row()
            if model_settings["checkpoint"] not in active_row.checkpoints_names:
                raise ValueError(
                    f"Checkpoint '{model_settings['checkpoint']}' not found in selected task"
                )

            active_row.set_selected_checkpoint_by_name(model_settings["checkpoint"])
        self.model_selector_cb()
        # ----------------------------------------- #

    def _init_hyperparameters(self, hyperparameters_settings: dict, options: dict) -> None:
        """
        Initialize the hyperparameters selector with the given settings.

        :param hyperparameters_settings: The hyperparameters settings.
        :type hyperparameters_settings: dict
        :param options: The application options.
        :type options: dict
        """
        self.hyperparameters_selector.set_hyperparameters(hyperparameters_settings)

        model_benchmark_settings = options.get("model_benchmark", None)
        if model_benchmark_settings is not None:
            self.hyperparameters_selector.set_model_benchmark_checkbox_value(
                model_benchmark_settings["enable"]
            )
            self.hyperparameters_selector.set_speedtest_checkbox_value(
                model_benchmark_settings["speed_test"]
                if model_benchmark_settings["enable"]
                else False
            )
        export_weights_settings = options.get("export", None)
        if export_weights_settings is not None:
            self.hyperparameters_selector.set_export_onnx_checkbox_value(
                export_weights_settings.get(RuntimeType.ONNXRUNTIME, False)
            )
            self.hyperparameters_selector.set_export_tensorrt_checkbox_value(
                export_weights_settings.get(RuntimeType.TENSORRT, False)
            )
        self.hyperparameters_selector_cb()

    # ----------------------------------------- #



import time
import uuid
from pathlib import Path
from queue import Queue
from threading import Event, Thread
from typing import Any, BinaryIO, Dict, List, Optional

import numpy as np
from pydantic import ValidationError

from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Geometry, Label
from supervisely.annotation.obj_class import ObjClass
from supervisely.api.api import Api
from supervisely.api.module_api import ApiField
from supervisely.api.video.video_figure_api import FigureInfo
from supervisely.geometry.helpers import deserialize_geometry
from supervisely.geometry.rectangle import Rectangle
from supervisely.imaging import image as sly_image
from supervisely.nn.inference.tracking.base_tracking import BaseTracking
from supervisely.nn.inference.tracking.tracker_interface import (
    TrackerInterface,
    TrackerInterfaceV2,
)
from supervisely.nn.prediction_dto import Prediction, PredictionBBox
from supervisely.sly_logger import logger
from supervisely.task.progress import Progress


class BBoxTracking(BaseTracking):
    def _deserialize_geometry(self, data: dict):
        geometry_type_str = data["type"]
        geometry_json = data["data"]
        return deserialize_geometry(geometry_type_str, geometry_json)

    def _track(self, api: Api, context: dict, notify_annotation_tool: bool):
        video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=False,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
            should_notify=notify_annotation_tool,
        )

        range_of_frames = [
            video_interface.frames_indexes[0],
            video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )

        api.logger.info("Start tracking.")

        def _upload_loop(q: Queue, stop_event: Event, video_interface: TrackerInterface):
            try:
                while True:
                    items = []
                    while not q.empty():
                        items.append(q.get_nowait())
                    if len(items) > 0:
                        video_interface.add_object_geometries_on_frames(*list(zip(*items)))
                        continue
                    if stop_event.is_set():
                        video_interface._notify(True, task="stop tracking")
                        return
                    time.sleep(1)
            except Exception as e:
                api.logger.error("Error in upload loop: %s", str(e), exc_info=True)
                video_interface._notify(True, task="stop tracking")
                video_interface.global_stop_indicatior = True
                raise

        upload_queue = Queue()
        stop_upload_event = Event()
        Thread(
            target=_upload_loop,
            args=[upload_queue, stop_upload_event, video_interface],
            daemon=True,
        ).start()

        try:
            for fig_id, obj_id in zip(
                video_interface.geometries.keys(),
                video_interface.object_ids,
            ):
                init = False
                for _ in video_interface.frames_loader_generator():
                    geom = video_interface.geometries[fig_id]
                    if not isinstance(geom, Rectangle):
                        stop_upload_event.set()
                        raise TypeError(f"Tracking does not work with {geom.geometry_name()}.")

                    imgs = video_interface.frames
                    target = PredictionBBox(
                        "",  # TODO: can this be useful?
                        [geom.top, geom.left, geom.bottom, geom.right],
                        None,
                    )

                    if not init:
                        self.initialize(imgs[0], target)
                        init = True

                    geometry = self.predict(
                        rgb_image=imgs[-1],
                        prev_rgb_image=imgs[0],
                        target_bbox=target,
                        settings=self.custom_inference_settings_dict,
                    )
                    sly_geometry = self._to_sly_geometry(geometry)
                    upload_queue.put(
                        (sly_geometry, obj_id, video_interface._cur_frames_indexes[-1])
                    )

                    if video_interface.global_stop_indicatior:
                        stop_upload_event.set()
                        return

                api.logger.info(f"Figure #{fig_id} tracked.")
        except Exception:
            stop_upload_event.set()
            raise
        stop_upload_event.set()

    def _track_api(self, api: Api, context: dict, request_uuid: str = None):
        track_t = time.monotonic()
        # unused fields:
        context["trackId"] = "auto"
        context["objectIds"] = []
        context["figureIds"] = []
        if "direction" not in context:
            context["direction"] = "forward"

        input_bboxes: list = context["input_geometries"]

        video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=False,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
            should_notify=False,
        )

        range_of_frames = [
            video_interface.frames_indexes[0],
            video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            # if cache is persistent, run cache task for whole video
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )

        predictions = []
        frames_n = video_interface.frames_count
        box_n = len(input_bboxes)
        geom_t = time.monotonic()
        api.logger.info(
            "Start tracking.",
            extra={
                "video_id": video_interface.video_id,
                "frame_range": range_of_frames,
                "geometries_count": box_n,
                "frames_count": frames_n,
                "request_uuid": request_uuid,
            },
        )
        for box_i, input_geom in enumerate(input_bboxes, 1):
            input_bbox = input_geom["data"]
            bbox = Rectangle.from_json(input_bbox)
            predictions_for_object = []
            init = False
            frame_t = time.monotonic()
            for frame_i, _ in enumerate(video_interface.frames_loader_generator(), 1):
                imgs = video_interface.frames
                target = PredictionBBox(
                    "",  # TODO: can this be useful?
                    [bbox.top, bbox.left, bbox.bottom, bbox.right],
                    None,
                )

                if not init:
                    self.initialize(imgs[0], target)
                    init = True

                geometry = self.predict(
                    rgb_image=imgs[-1],
                    prev_rgb_image=imgs[0],
                    target_bbox=target,
                    settings=self.custom_inference_settings_dict,
                )
                sly_geometry = self._to_sly_geometry(geometry)

                predictions_for_object.append(
                    {"type": sly_geometry.geometry_name(), "data": sly_geometry.to_json()}
                )
                api.logger.debug(
                    "Frame processed. Geometry: [%d / %d]. Frame: [%d / %d]",
                    box_i,
                    box_n,
                    frame_i,
                    frames_n,
                    extra={
                        "geometry_index": box_i,
                        "frame_index": frame_i,
                        "processing_time": time.monotonic() - frame_t,
                        "request_uuid": request_uuid,
                    },
                )
                frame_t = time.monotonic()

            predictions.append(predictions_for_object)
            api.logger.info(
                "Geometry processed. Progress: [%d / %d]",
                box_i,
                box_n,
                extra={
                    "geometry_index": box_i,
                    "processing_time": time.monotonic() - geom_t,
                    "request_uuid": request_uuid,
                },
            )
            geom_t = time.monotonic()

        # predictions must be NxK bboxes: N=number of frames, K=number of objects
        predictions = list(map(list, zip(*predictions)))
        api.logger.info(
            "Tracking finished.",
            extra={"tracking_time": time.monotonic() - track_t, "request_uuid": request_uuid},
        )
        return predictions

    def _inference(self, frames: List[np.ndarray], geometries: List[Geometry], settings: dict):
        updated_settings = {
            **self.custom_inference_settings_dict,
            **settings,
        }
        results = [[] for _ in range(len(frames) - 1)]
        for geometry in geometries:
            if not isinstance(geometry, Rectangle):
                raise TypeError(f"Tracking does not work with {geometry.geometry_name()}.")
            target = PredictionBBox(
                "",
                [geometry.top, geometry.left, geometry.bottom, geometry.right],
                None,
            )
            self.initialize(frames[0], target)
            for i in range(len(frames) - 1):
                pred_geometry = self.predict(
                    rgb_image=frames[i + 1],
                    prev_rgb_image=frames[i],
                    target_bbox=target,
                    settings=updated_settings,
                )
                sly_pred_geometry = self._to_sly_geometry(pred_geometry)
                results[i].append(
                    {"type": Rectangle.geometry_name(), "data": sly_pred_geometry.to_json()}
                )
        return results

    def _track_async(self, api: Api, context: dict, inference_request_uuid: str = None):
        inference_request = self._inference_requests[inference_request_uuid]
        tracker_interface = TrackerInterfaceV2(api, context, self.cache)
        progress: Progress = inference_request["progress"]
        frames_count = tracker_interface.frames_count
        figures = tracker_interface.figures
        progress_total = frames_count * len(figures)
        progress.total = progress_total

        def _upload_f(items: List[FigureInfo]):
            with inference_request["lock"]:
                inference_request["pending_results"].extend(items)

        def _notify_f(items: List[FigureInfo]):
            items_by_object_id: Dict[int, List[FigureInfo]] = {}
            for item in items:
                items_by_object_id.setdefault(item.object_id, []).append(item)

            for object_id, object_items in items_by_object_id.items():
                frame_range = [
                    min(item.frame_index for item in object_items),
                    max(item.frame_index for item in object_items),
                ]
                progress.iters_done_report(len(object_items))
                tracker_interface.notify_progress(progress.current, progress.total, frame_range)

        api.logger.info("Start tracking.")
        try:
            with tracker_interface(_upload_f, _notify_f):
                for fig_i, figure in enumerate(figures, 1):
                    figure = api.video.figure._convert_json_info(figure)
                    if not figure.geometry_type == Rectangle.geometry_name():
                        raise TypeError(f"Tracking does not work with {figure.geometry_type}.")
                    api.logger.info("figure:", extra={"figure": figure._asdict()})
                    sly_geometry: Rectangle = deserialize_geometry(
                        figure.geometry_type, figure.geometry
                    )
                    init = False
                    for frame_i, (frame, next_frame) in enumerate(
                        tracker_interface.frames_loader_generator(), 1
                    ):
                        target = PredictionBBox(
                            "",  # TODO: can this be useful?
                            [
                                sly_geometry.top,
                                sly_geometry.left,
                                sly_geometry.bottom,
                                sly_geometry.right,
                            ],
                            None,
                        )

                        if not init:
                            self.initialize(frame.image, target)
                            init = True

                        logger.debug("Start prediction")
                        t = time.time()
                        geometry = self.predict(
                            rgb_image=next_frame.image,
                            prev_rgb_image=frame.image,
                            target_bbox=target,
                            settings=self.custom_inference_settings_dict,
                        )
                        logger.debug("Prediction done. Time: %f sec", time.time() - t)
                        sly_geometry = self._to_sly_geometry(geometry)

                        figure_id = uuid.uuid5(
                            namespace=uuid.NAMESPACE_URL, name=f"{time.time()}"
                        ).hex
                        result_figure = api.video.figure._convert_json_info(
                            {
                                ApiField.ID: figure_id,
                                ApiField.OBJECT_ID: figure.object_id,
                                "meta": {"frame": next_frame.frame_index},
                                ApiField.GEOMETRY_TYPE: sly_geometry.geometry_name(),
                                ApiField.GEOMETRY: sly_geometry.to_json(),
                                ApiField.TRACK_ID: tracker_interface.track_id,
                            }
                        )

                        tracker_interface.add_prediction(result_figure)

                        logger.debug(
                            "Frame [%d / %d] processed.",
                            frame_i,
                            tracker_interface.frames_count,
                        )

                        if inference_request["cancel_inference"]:
                            return
                        if tracker_interface.is_stopped():
                            reason = tracker_interface.stop_reason()
                            if isinstance(reason, Exception):
                                raise reason
                            return

                    api.logger.info(
                        "Figure [%d, %d] tracked.",
                        fig_i,
                        len(figures),
                        extra={"figure_id": figure.id},
                    )
        except Exception:
            progress.message = "Error occured during tracking"
            raise
        else:
            progress.message = "Ready"
        finally:
            progress.set(current=0, total=1, report=True)

    def track(self, api: Api, state: Dict, context: Dict):
        fn = self.send_error_data(api, context)(self._track)
        self.schedule_task(fn, api, context, notify_annotation_tool=True)
        return {"message": "Track task started."}

    def track_api(self, api: Api, state: Dict, context: Dict):
        request_uuid = uuid.uuid5(namespace=uuid.NAMESPACE_URL, name=f"{time.time()}").hex
        result = self._track_api(api, context, request_uuid)
        logger.info("Track-api request processed.", extra={"request_uuid": request_uuid})
        return result

    def track_api_files(self, files: List[BinaryIO], settings: Dict):
        logger.info("Start tracking with settings:", extra={"settings": settings})
        frame_indexes = list(
            range(settings["frame_index"], settings["frame_index"] + settings["frames"] + 1)
        )
        geometries = map(self._deserialize_geometry, settings["input_geometries"])
        frames = []
        for file, frame_idx in zip(files, frame_indexes):
            img_bytes = file.read()
            frame = sly_image.read_bytes(img_bytes)
            frames.append(frame)
        logger.info("Start tracking.")
        return self._inference(frames, geometries, settings)

    def track_async(self, api: Api, state: Dict, context: Dict):
        batch_size = context.get("batch_size", self.get_batch_size())
        if self.max_batch_size is not None and batch_size > self.max_batch_size:
            raise ValidationError(
                f"Batch size should be less than or equal to {self.max_batch_size} for this model."
            )

        inference_request_uuid = uuid.uuid5(namespace=uuid.NAMESPACE_URL, name=f"{time.time()}").hex
        fn = self.send_error_data(api, context)(self._track_async)
        self.schedule_task(fn, api, context, inference_request_uuid=inference_request_uuid)

        logger.debug(
            "Inference has scheduled from 'track_async' endpoint",
            extra={"inference_request_uuid": inference_request_uuid},
        )
        return {
            "message": "Inference has started.",
            "inference_request_uuid": inference_request_uuid,
        }

    def initialize(self, init_rgb_image: np.ndarray, target_bbox: PredictionBBox) -> None:
        """
        Initializing the tracker with a new object.

        :param init_rgb_image: frame with object
        :type init_rgb_image: np.ndarray
        :param target_bbox: initial bbox
        :type target_bbox: PredictionBBox
        """
        raise NotImplementedError

    def predict(
        self,
        rgb_image: np.ndarray,
        settings: Dict[str, Any],
        prev_rgb_image: np.ndarray,
        target_bbox: PredictionBBox,
    ) -> PredictionBBox:
        """
        SOT prediction

        :param rgb_image: search frame
        :type rgb_image: np.ndarray
        :param settings: model parameters
        :type settings: Dict[str, Any]
        :param init_rgb_image: previous frame with object
        :type init_rgb_image: np.ndarray
        :param target_bbox: bbox added on previous step
        :type target_bbox: PredictionBBox
        :return: predicted annotation
        :rtype: PredictionBBox
        """
        raise NotImplementedError

    def visualize(
        self,
        predictions: List[PredictionBBox],
        images: List[np.ndarray],
        vis_path: str,
        thickness: int = 2,
        classes_whitelist: Optional[List[str]] = None,
    ):
        vis_path = Path(vis_path)

        for i, (pred, image) in enumerate(zip(predictions, images)):
            out_path = vis_path / f"img_{i}.jpg"
            ann = self._predictions_to_annotation(image, [pred], classes_whitelist)
            ann.draw_pretty(
                bitmap=image,
                color=(255, 0, 0),
                thickness=thickness,
                output_path=str(out_path),
                fill_rectangles=False,
            )

    def _to_sly_geometry(self, dto: PredictionBBox) -> Rectangle:
        top, left, bottom, right = dto.bbox_tlbr
        geometry = Rectangle(top=top, left=left, bottom=bottom, right=right)
        return geometry

    def _create_label(self, dto: PredictionBBox) -> Rectangle:
        geometry = self._to_sly_geometry(dto)
        return Label(geometry, ObjClass("", Rectangle))

    def _get_obj_class_shape(self):
        return Rectangle

    def _predictions_to_annotation(
        self,
        image: np.ndarray,
        predictions: List[Prediction],
        classes_whitelist: Optional[List[str]] = None,
    ) -> Annotation:
        labels = []
        for prediction in predictions:
            if (
                not classes_whitelist in (None, "all")
                and prediction.class_name not in classes_whitelist
            ):
                continue
            label = self._create_label(prediction)
            if label is None:
                # for example empty mask
                continue
            if isinstance(label, list):
                labels.extend(label)
                continue
            labels.append(label)

        # create annotation with correct image resolution
        ann = Annotation(img_size=image.shape[:2])
        ann = ann.add_labels(labels)
        return ann



import os
from typing import List, Tuple
import numpy as np
from supervisely import logger, fs
from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d

TABLE_NAMES = [
    "category",
    "attribute",
    "visibility",
    "instance",
    "sensor",
    "calibrated_sensor",
    "ego_pose",
    "log",
    "scene",
    "sample",
    "sample_data",
    "sample_annotation",
    "map",
]
FOLDER_NAMES = ["data", "lidar", "images", "maps"]


def get_available_scenes(lyft):
    for scene in lyft.scene:
        token = scene["token"]
        scene_rec = lyft.get("scene", token)
        sample_rec = lyft.get("sample", scene_rec["first_sample_token"])
        sample_data = lyft.get("sample_data", sample_rec["data"]["LIDAR_TOP"])

        lidar_path, boxes, _ = lyft.get_sample_data(sample_data["token"])
        if not os.path.exists(str(lidar_path)):
            continue
        yield scene


def extract_data_from_scene(lyft, scene):
    try:
        from pyquaternion import Quaternion
        from lyft_dataset_sdk.utils.geometry_utils import transform_matrix
    except ImportError as ie:
        logger.warn(f"Lazy import failed. Error: {ie}")
        return

    new_token = scene["first_sample_token"]
    dataset_data = []
    num_samples = scene["nbr_samples"]
    for i in range(num_samples):
        my_sample = lyft.get("sample", new_token)

        data = {}
        data["ann_data"] = {}
        data["ann_data"]["timestamp"] = my_sample["timestamp"]
        data["custom_data"] = lyft.get("log", scene["log_token"])

        sensor_token = my_sample["data"]["LIDAR_TOP"]
        lidar_path, boxes, _ = lyft.get_sample_data(sensor_token)
        if not os.path.exists(str(lidar_path)):
            logger.debug(f"Skipping sample {new_token} - lidar file doesn't exist")
            continue
        data["lidar_path"] = str(lidar_path)

        sd_record_lid = lyft.get("sample_data", sensor_token)
        cs_record_lid = lyft.get("calibrated_sensor", sd_record_lid["calibrated_sensor_token"])
        ego_record_lid = lyft.get("ego_pose", sd_record_lid["ego_pose_token"])

        locs = np.array([b.center for b in boxes]).reshape(-1, 3)
        dims = np.array([b.wlh for b in boxes]).reshape(-1, 3)
        rots = np.array([b.orientation.yaw_pitch_roll[0] for b in boxes]).reshape(-1, 1)

        gt_boxes = np.concatenate([locs, dims, -rots + np.pi / 2], axis=1)
        names = np.array([b.name for b in boxes])
        data["ann_data"]["names"] = names
        data["ann_data"]["gt_boxes"] = gt_boxes
        instance_tokens = [lyft.get("sample_annotation", box.token) for box in boxes]
        data["ann_data"]["instance_tokens"] = instance_tokens

        for sensor, sensor_token in my_sample["data"].items():
            if "CAM" in sensor:
                img_path, boxes, cam_intrinsic = lyft.get_sample_data(sensor_token)
                if not os.path.exists(img_path):
                    logger.debug(f"pass {sensor} - image doesn't exist")
                    continue
                data["ann_data"][sensor] = str(img_path)

                sd_record_cam = lyft.get("sample_data", sensor_token)
                cs_record_cam = lyft.get(
                    "calibrated_sensor", sd_record_cam["calibrated_sensor_token"]
                )
                ego_record_cam = lyft.get("ego_pose", sd_record_cam["ego_pose_token"])
                cam_height = sd_record_cam["height"]
                cam_width = sd_record_cam["width"]

                lid_to_ego = transform_matrix(
                    cs_record_lid["translation"],
                    Quaternion(cs_record_lid["rotation"]),
                    inverse=False,
                )
                lid_ego_to_world = transform_matrix(
                    ego_record_lid["translation"],
                    Quaternion(ego_record_lid["rotation"]),
                    inverse=False,
                )
                world_to_cam_ego = transform_matrix(
                    ego_record_cam["translation"],
                    Quaternion(ego_record_cam["rotation"]),
                    inverse=True,
                )
                ego_to_cam = transform_matrix(
                    cs_record_cam["translation"],
                    Quaternion(cs_record_cam["rotation"]),
                    inverse=True,
                )

                velo_to_cam = np.dot(
                    ego_to_cam, np.dot(world_to_cam_ego, np.dot(lid_ego_to_world, lid_to_ego))
                )
                velo_to_cam_rot = velo_to_cam[:3, :3]
                velo_to_cam_trans = velo_to_cam[:3, 3]

                data["ann_data"][f"{sensor}_extrinsic"] = np.hstack(
                    (velo_to_cam_rot, velo_to_cam_trans.reshape(3, 1))
                )
                data["ann_data"][f"{sensor}_intrinsic"] = np.asarray(
                    cs_record_cam["camera_intrinsic"]
                )
                data["ann_data"][f"{sensor}_imsize"] = (cam_width, cam_height)
            else:
                logger.debug(f"pass {sensor} - isn't a camera")
        dataset_data.append(data)

        new_token = my_sample.get("next")
        if not new_token:
            break

    return dataset_data


def generate_rimage_infos(related_images: List[Tuple[str, str]], ann_data):
    sensors_to_skip = ["_intrinsic", "_extrinsic", "_imsize"]
    for sensor, image_path in related_images:
        if not any([sensor.endswith(s) for s in sensors_to_skip]):
            image_name = fs.get_file_name_with_ext(image_path)
            sly_path_img = os.path.join(os.path.dirname(image_path), image_name)
            img_info = {
                "name": image_name,
                "meta": {
                    "deviceId": sensor,
                    "timestamp": ann_data["timestamp"],
                    "sensorsData": {
                        "extrinsicMatrix": list(
                            ann_data[f"{sensor}_extrinsic"].flatten().astype(float)
                        ),
                        "intrinsicMatrix": list(
                            ann_data[f"{sensor}_intrinsic"].flatten().astype(float)
                        ),
                    },
                },
            }
            yield sly_path_img, img_info


def _convert_BEVBox3D_to_geometry(box):
    bbox = box.to_xyzwhlr()
    dim = bbox[[3, 5, 4]]
    pos = bbox[:3] + [0, 0, dim[1] / 2]
    yaw = bbox[-1]
    position = Vector3d(float(pos[0]), float(pos[1]), float(pos[2]))
    rotation = Vector3d(0, 0, float(-yaw))

    dimension = Vector3d(float(dim[0]), float(dim[2]), float(dim[1]))
    geometry = Cuboid3d(position, rotation, dimension)
    return geometry


def convert_bin_to_pcd(bin_file, save_filepath):
    import open3d as o3d  # pylint: disable=import-error

    b = np.fromfile(bin_file, dtype=np.float32).reshape(-1, 5)
    points = b[:, 0:3]
    intensity = b[:, 3]
    ring_index = b[:, 4]
    intensity_fake_rgb = np.zeros((intensity.shape[0], 3))
    intensity_fake_rgb[:, 0] = (
        intensity  # red The intensity measures the reflectivity of the objects
    )
    intensity_fake_rgb[:, 1] = (
        ring_index  # green ring index is the index of the laser ranging from 0 to 31
    )

    pc = o3d.geometry.PointCloud(o3d.utility.Vector3dVector(points))
    pc.colors = o3d.utility.Vector3dVector(intensity_fake_rgb)
    o3d.io.write_point_cloud(save_filepath, pc)


def get_related_images(ann_data):
    # sensors_to_skip = ["_intrinsic", "_extrinsic", "_imsize"]
    # return [
    #     (sensor, img_path)
    #     for sensor, img_path in ann_data.items()
    #     if "CAM" in sensor and not any([sensor.endswith(s) for s in sensors_to_skip])
    # ]
    return [(sensor, img_path) for sensor, img_path in ann_data.items() if "CAM" in sensor]


def validate_ann_dir(ann_dir):
    if any([not fs.file_exists(f"{ann_dir}/{d}.json") for d in TABLE_NAMES]):
        return False
    return True


def lyft_annotation_to_BEVBox3D(data):
    import open3d as o3d  # pylint: disable=import-error

    boxes = data["gt_boxes"]
    names = data["names"]

    objects = []
    for name, box in zip(names, boxes):
        center = [float(box[0]), float(box[1]), float(box[2])]
        size = [float(box[3]), float(box[5]), float(box[4])]
        ry = float(box[6])

        yaw = ry - np.pi
        yaw = yaw - np.floor(yaw / (2 * np.pi) + 0.5) * 2 * np.pi
        world_cam = None
        objects.append(o3d.ml.datasets.utils.BEVBox3D(center, size, yaw, name, -1.0, world_cam))
        objects[-1].yaw = ry

    return objects



import pickle
from pathlib import Path
from typing import Dict, List, Optional

import pandas as pd

from supervisely.annotation.annotation import ProjectMeta
from supervisely.api.api import Api
from supervisely.api.module_api import ApiField
from supervisely.api.dataset_api import DatasetInfo
from supervisely.api.project_api import ProjectInfo
from supervisely.app.widgets import SlyTqdm
from supervisely.io.env import team_id
from supervisely.io.fs import dir_empty, mkdir
from supervisely.io.json import load_json_file
from supervisely.nn.benchmark.object_detection.metric_provider import MetricProvider
from supervisely.nn.benchmark.visualization.vis_click_data import ClickData, IdMapper
from supervisely.sly_logger import logger
from supervisely.task.progress import tqdm_sly

# class ImageComparisonData:
#     def __init__(
#         self,
#         gt_image_info: ImageInfo = None,
#         pred_image_info: ImageInfo = None,
#         diff_image_info: ImageInfo = None,
#         gt_annotation: Annotation = None,
#         pred_annotation: Annotation = None,
#         diff_annotation: Annotation = None,
#     ):
#         self.gt_image_info = gt_image_info
#         self.pred_image_info = pred_image_info
#         self.diff_image_info = diff_image_info
#         self.gt_annotation = gt_annotation
#         self.pred_annotation = pred_annotation
#         self.diff_annotation = diff_annotation


class EvalResult:

    def __init__(
        self,
        eval_dir: str,
        workdir: str,
        api: Api,
        progress: Optional[SlyTqdm] = None,
    ):
        from pycocotools.coco import COCO  # pylint: disable=import-error

        self.eval_dir = eval_dir
        self.report_path = Path(eval_dir, "visualizations", "template.vue").as_posix()
        self.workdir = workdir
        self.api = api
        self.team_id = team_id()
        self.local_dir = str(Path(self.workdir, self.eval_dir.lstrip("/")))
        self.progress = progress or tqdm_sly

        self.coco_gt: COCO = None
        self.coco_dt: COCO = None
        self.inference_info: Dict = None
        self.speedtest_info: Dict = None
        self.eval_data: Dict = None
        self.mp: MetricProvider = None
        self.df_score_profile: pd.DataFrame = None
        self.dfsp_down: pd.DataFrame = None
        self.f1_optimal_conf: float = None
        self.click_data: ClickData = None
        # self.comparison_data: Dict[int, ImageComparisonData] = {}
        self.color = None

        self._gt_project_info = None
        self._gt_project_meta = None
        self._gt_dataset_infos = None
        self._dt_project_id = None
        self._dt_project_meta = None

        self._load_eval_data()
        self._read_eval_data()

    @property
    def cv_task(self):
        return self.inference_info.get("task_type")

    @property
    def name(self) -> str:
        model_name = self.inference_info.get("model_name", self.eval_dir)
        return self.inference_info.get("deploy_params", {}).get("checkpoint_name", model_name)

    @property
    def model_name(self) -> str:
        if not self.name:
            return
        if len(self.name) > 20:
            return self.name[:9] + "..." + self.name[-6:]
        return self.name

    @property
    def gt_project_id(self) -> int:
        return self.inference_info.get("gt_project_id")

    @property
    def gt_project_info(self) -> ProjectInfo:
        if self._gt_project_info is None:
            gt_project_id = self.inference_info.get("gt_project_id")
            self._gt_project_info = self.api.project.get_info_by_id(gt_project_id)
        return self._gt_project_info

    @property
    def gt_project_meta(self) -> ProjectMeta:
        if self._gt_project_meta is None:
            self._gt_project_meta = ProjectMeta.from_json(
                self.api.project.get_meta(self.gt_project_id)
            )
        return self._gt_project_meta

    @property
    def gt_dataset_ids(self) -> List[int]:
        return self.inference_info.get("gt_dataset_ids", None)

    @property
    def gt_dataset_infos(self) -> List[DatasetInfo]:
        if self._gt_dataset_infos is None:
            filters = None
            if self.gt_dataset_ids is not None:
                filters = [
                    {
                        ApiField.FIELD: ApiField.ID,
                        ApiField.OPERATOR: "in",
                        ApiField.VALUE: self.gt_dataset_ids,
                    }
                ]
            self._gt_dataset_infos = self.api.dataset.get_list(
                self.gt_project_id,
                filters=filters,
                recursive=True,
            )
        return self._gt_dataset_infos

    @property
    def dt_project_id(self):
        if self._dt_project_id is None:
            self._dt_project_id = self.inference_info.get("dt_project_id")
        return self._dt_project_id

    @property
    def dt_project_meta(self):
        if self._dt_project_meta is None:
            self._dt_project_meta = ProjectMeta.from_json(
                self.api.project.get_meta(self.dt_project_id)
            )
        return self._dt_project_meta

    @property
    def train_info(self):
        return self.inference_info.get("train_info", None)  # TODO: check

    @property
    def gt_images_ids(self):
        return self.inference_info.get("gt_images_ids", None)  # TODO: check

    @property
    def classes_whitelist(self):
        return self.inference_info.get("inference_settings", {}).get("classes", [])  # TODO: check

    def _load_eval_data(self):
        if not dir_empty(self.local_dir):
            return
        if not self.api.storage.dir_exists(self.team_id, self.eval_dir):
            raise ValueError(f"Directory {self.eval_dir} not found in storage.")
        mkdir(self.local_dir)
        dir_name = Path(self.eval_dir).name
        with self.progress(
            message=f"Downloading evaluation data from {dir_name}",
            total=self.api.storage.get_directory_size(self.team_id, self.eval_dir),
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as pbar:
            self.api.storage.download_directory(
                self.team_id, self.eval_dir, self.local_dir, progress_cb=pbar.update
            )

    # def _load_projects(self):
    #     projects_dir = Path(self.local_dir, "projects")
    #     items_total = self.gt_images_ids
    #     if items_total is None:
    #         items_total = sum(self.gt_dataset_infos, key=lambda x: x.items_count)
    #     with self.progress(
    #         message=f"Downloading GT project {self.gt_project_info.name} and datasets",
    #         total=items_total,
    #     ) as pbar:
    #         download_project(
    #             self.api,
    #             self.gt_project_info.id,
    #             str(projects_dir),
    #             dataset_ids=self.gt_dataset_ids,
    #             progress_cb=pbar.update,
    #         )

    def _read_eval_data(self):
        from pycocotools.coco import COCO  # pylint: disable=import-error

        gt_path = str(Path(self.local_dir, "evaluation", "cocoGt.json"))
        dt_path = str(Path(self.local_dir, "evaluation", "cocoDt.json"))
        coco_gt, coco_dt = COCO(gt_path), COCO(dt_path)
        self.coco_gt = coco_gt
        self.coco_dt = coco_dt
        self.eval_data = pickle.load(
            open(Path(self.local_dir, "evaluation", "eval_data.pkl"), "rb")
        )
        self.inference_info = load_json_file(
            Path(self.local_dir, "evaluation", "inference_info.json")
        )
        speedtest_info_path = Path(self.local_dir, "speedtest", "speedtest.json")
        if speedtest_info_path.exists():
            self.speedtest_info = load_json_file(
                Path(self.local_dir, "speedtest", "speedtest.json")
            )

        self.mp = MetricProvider(
            self.eval_data,
            self.coco_gt,
            self.coco_dt,
        )
        self.mp.calculate()

        self.df_score_profile = pd.DataFrame(
            self.mp.confidence_score_profile(), columns=["scores", "precision", "recall", "f1"]
        )

        # downsample
        if len(self.df_score_profile) > 5000:
            self.dfsp_down = self.df_score_profile.iloc[:: len(self.df_score_profile) // 1000]
        else:
            self.dfsp_down = self.df_score_profile

        self.f1_optimal_conf = self.mp.get_f1_optimal_conf()[0]
        if self.f1_optimal_conf is None:
            self.f1_optimal_conf = 0.01
            logger.warning("F1 optimal confidence cannot be calculated. Using 0.01 as default.")

        # Click data
        gt_id_mapper = IdMapper(self.coco_gt.dataset)
        dt_id_mapper = IdMapper(self.coco_dt.dataset)

        self.click_data = ClickData(self.mp.m, gt_id_mapper, dt_id_mapper)

    # def _update_comparison_data(
    #     self,
    #     gt_image_id: int,
    #     gt_image_info: ImageInfo = None,
    #     pred_image_info: ImageInfo = None,
    #     diff_image_info: ImageInfo = None,
    #     gt_annotation: Annotation = None,
    #     pred_annotation: Annotation = None,
    #     diff_annotation: Annotation = None,
    # ):
    #     comparison_data = self.comparison_data.get(gt_image_id, None)
    #     if comparison_data is None:
    #         self.comparison_data[gt_image_id] = ImageComparisonData(
    #             gt_image_info=gt_image_info,
    #             pred_image_info=pred_image_info,
    #             diff_image_info=diff_image_info,
    #             gt_annotation=gt_annotation,
    #             pred_annotation=pred_annotation,
    #             diff_annotation=diff_annotation,
    #         )
    #     else:
    #         for attr, value in {
    #             "gt_image_info": gt_image_info,
    #             "pred_image_info": pred_image_info,
    #             "diff_image_info": diff_image_info,
    #             "gt_annotation": gt_annotation,
    #             "pred_annotation": pred_annotation,
    #             "diff_annotation": diff_annotation,
    #         }.items():
    #             if value is not None:
    #                 setattr(comparison_data, attr, value)



# coding: utf-8

# docs
from typing import Callable, Dict, List, Optional, Tuple, Union

from tqdm import tqdm

from supervisely._utils import batched
from supervisely.api.module_api import ApiField, ModuleApiBase


class AdvancedApi(ModuleApiBase):
    """class AdvancedApi"""

    def add_tag_to_object(
        self, tag_meta_id: int, figure_id: int, value: Optional[Union[str, int]] = None
    ) -> Dict:
        """add_tag_to_object"""
        data = {ApiField.TAG_ID: tag_meta_id, ApiField.FIGURE_ID: figure_id}
        if value is not None:
            data[ApiField.VALUE] = value
        resp = self._api.post("object-tags.add-to-object", data)
        return resp.json()

    def remove_tag_from_object(self, tag_meta_id: int, figure_id: int, tag_id: int) -> Dict:
        """remove_tag_from_object"""
        data = {ApiField.TAG_ID: tag_meta_id, ApiField.FIGURE_ID: figure_id, ApiField.ID: tag_id}
        resp = self._api.post("object-tags.remove-from-figure", data)
        return resp.json()

    def get_object_tags(self, figure_id: int) -> Dict:
        """get_object_tags"""
        data = {ApiField.ID: figure_id}
        resp = self._api.post("figures.tags.list", data)
        return resp.json()

    def remove_tag_from_image(self, tag_meta_id: int, image_id: int, tag_id: int) -> Dict:
        """remove_tag_from_image"""
        data = {ApiField.TAG_ID: tag_meta_id, ApiField.IMAGE_ID: image_id, ApiField.ID: tag_id}
        resp = self._api.post("image-tags.remove-from-image", data)
        return resp.json()

    def remove_tags_from_images(
        self,
        tag_meta_ids: List[int],
        image_ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> None:
        """remove_tags_from_images"""
        for batch_ids in batched(image_ids, batch_size=100):
            data = {ApiField.TAG_IDS: tag_meta_ids, ApiField.IDS: batch_ids}
            self._api.post("image-tags.bulk.remove-from-images", data)
            if progress_cb is not None:
                progress_cb(len(batch_ids))



from supervisely.cli.cli import cli
import supervisely as sly
from typing import Callable, Literal, Union


def _handle_creds_error_to_console(
    api_func: sly.Api, printer: Callable = print
) -> Union[sly.Api, Literal[False]]:
    try:
        api = api_func()
    except (KeyError, ValueError) as e:
        if "server_address is not defined" in repr(e):
            text = "SERVER_ADDRESS env variable is undefined. Add it to your '~/supervisely.env' file or to environment variables. Learn more here: https://developer.supervise.ly/getting-started/basics-of-authentication."
            printer(f"\n{text}\n", style="bold red") if "Console.print" in repr(
                printer
            ) else printer(text)
            return False
        elif "api_token is not defined" in repr(e):
            text = "API_TOKEN env variable is undefined. Add it to your '~/supervisely.env' file or to environment variables. Learn more here: https://developer.supervise.ly/getting-started/basics-of-authentication."
            printer(f"\n{text}\n", style="bold red") if "Console.print" in repr(
                printer
            ) else printer(text)
            return False
        else:
            printer(repr(e))
            return False
    return api
